@@ -15,29 +15,59 @@
 
 import collections
 import os
+import threading
 
 import eventlet
+from neutron_lib.agent import constants as agent_consts
+from neutron_lib.agent import topics
 from neutron_lib import constants
+from neutron_lib import context
 from neutron_lib import exceptions
+from neutron_lib import rpc as n_rpc
+from oslo_concurrency import lockutils
 from oslo_config import cfg
 from oslo_log import log as logging
 import oslo_messaging
 from oslo_service import loopingcall
+from oslo_utils import fileutils
 from oslo_utils import importutils
+import six
 
-from neutron._i18n import _, _LE, _LI, _LW
+from neutron._i18n import _
+from neutron.agent.common import resource_processing_queue as queue
 from neutron.agent.linux import dhcp
 from neutron.agent.linux import external_process
 from neutron.agent.metadata import driver as metadata_driver
 from neutron.agent import rpc as agent_rpc
 from neutron.common import constants as n_const
-from neutron.common import rpc as n_rpc
-from neutron.common import topics
 from neutron.common import utils
-from neutron import context
 from neutron import manager
 
 LOG = logging.getLogger(__name__)
+_SYNC_STATE_LOCK = lockutils.ReaderWriterLock()
+
+DEFAULT_PRIORITY = 255
+
+DHCP_PROCESS_GREENLET_MAX = 32
+DHCP_PROCESS_GREENLET_MIN = 8
+
+
+def _sync_lock(f):
+    """Decorator to block all operations for a global sync call."""
+    @six.wraps(f)
+    def wrapped(*args, **kwargs):
+        with _SYNC_STATE_LOCK.write_lock():
+            return f(*args, **kwargs)
+    return wrapped
+
+
+def _wait_if_syncing(f):
+    """Decorator to wait if any sync operations are in progress."""
+    @six.wraps(f)
+    def wrapped(*args, **kwargs):
+        with _SYNC_STATE_LOCK.read_lock():
+            return f(*args, **kwargs)
+    return wrapped
 
 
 class DhcpAgent(manager.Manager):
@@ -47,7 +77,7 @@ class DhcpAgent(manager.Manager):
     of an rpc interface.  The neutron server uses
     neutron.api.rpc.agentnotifiers.dhcp_rpc_agent_api.DhcpAgentNotifyApi as the
     client side to execute the methods here.  For more information about
-    changing rpc interfaces, see doc/source/devref/rpc_api.rst.
+    changing rpc interfaces, see doc/source/contributor/internals/rpc_api.rst.
     """
     target = oslo_messaging.Target(version='1.0')
 
@@ -56,12 +86,21 @@ class DhcpAgent(manager.Manager):
         self.needs_resync_reasons = collections.defaultdict(list)
         self.dhcp_ready_ports = set()
         self.conf = conf or cfg.CONF
+        # If 'resync_throttle' is configured more than 'resync_interval' by
+        # mistake, raise exception and log with message.
+        if self.conf.resync_throttle > self.conf.resync_interval:
+            msg = _("DHCP agent must have resync_throttle <= resync_interval")
+            LOG.exception(msg)
+            raise exceptions.InvalidConfigurationOption(
+                opt_name='resync_throttle',
+                opt_value=self.conf.resync_throttle)
+        self._periodic_resync_event = threading.Event()
         self.cache = NetworkCache()
         self.dhcp_driver_cls = importutils.import_class(self.conf.dhcp_driver)
         self.plugin_rpc = DhcpPluginApi(topics.PLUGIN, self.conf.host)
         # create dhcp dir to store dhcp info
         dhcp_dir = os.path.dirname("/%s/dhcp/" % self.conf.state_path)
-        utils.ensure_dir(dhcp_dir)
+        fileutils.ensure_tree(dhcp_dir, mode=0o755)
         self.dhcp_version = self.dhcp_driver_cls.check_version()
         self._populate_networks_cache()
         # keep track of mappings between networks and routers for
@@ -70,6 +109,9 @@ class DhcpAgent(manager.Manager):
         self._process_monitor = external_process.ProcessMonitor(
             config=self.conf,
             resource_type='dhcp')
+        self._pool_size = DHCP_PROCESS_GREENLET_MIN
+        self._pool = eventlet.GreenPool(size=self._pool_size)
+        self._queue = queue.ResourceProcessingQueue()
 
     def init_host(self):
         self.sync_state()
@@ -81,7 +123,8 @@ class DhcpAgent(manager.Manager):
                 self.conf
             )
             for net_id in existing_networks:
-                net = dhcp.NetModel({"id": net_id, "subnets": [], "ports": []})
+                net = dhcp.NetModel({"id": net_id, "subnets": [],
+                                     "non_local_subnets": [], "ports": []})
                 self.cache.put(net)
         except NotImplementedError:
             # just go ahead with an empty networks cache
@@ -91,13 +134,13 @@ class DhcpAgent(manager.Manager):
 
     def after_start(self):
         self.run()
-        LOG.info(_LI("DHCP agent started"))
+        LOG.info("DHCP agent started")
 
     def run(self):
         """Activate the DHCP agent."""
-        self.sync_state()
         self.periodic_resync()
         self.start_ready_ports_loop()
+        eventlet.spawn_n(self._process_loop)
 
     def call_driver(self, action, network, **action_kwargs):
         """Invoke an action on a DHCP driver instance."""
@@ -131,12 +174,13 @@ class DhcpAgent(manager.Manager):
                 # allocation pool or a port is  deleted to free up an IP, this
                 # will automatically be retried on the notification
                 self.schedule_resync(e, network.id)
-            if (isinstance(e, oslo_messaging.RemoteError)
-                and e.exc_type == 'NetworkNotFound'
-                or isinstance(e, exceptions.NetworkNotFound)):
-                LOG.debug("Network %s has been deleted.", network.id)
+            if (isinstance(e, oslo_messaging.RemoteError) and
+                    e.exc_type == 'NetworkNotFound' or
+                    isinstance(e, exceptions.NetworkNotFound)):
+                LOG.debug("Network %s has been removed from the agent "
+                          "or deleted from DB.", network.id)
             else:
-                LOG.exception(_LE('Unable to %(action)s dhcp for %(net_id)s.'),
+                LOG.exception('Unable to %(action)s dhcp for %(net_id)s.',
                               {'net_id': network.id, 'action': action})
 
     def schedule_resync(self, reason, network_id=None):
@@ -144,28 +188,35 @@ class DhcpAgent(manager.Manager):
         specified, resync all networks.
         """
         self.needs_resync_reasons[network_id].append(reason)
+        self._periodic_resync_event.set()
+        # Yield to allow other threads that may be ready to run.
+        # This helps prevent one thread from acquiring the same lock over and
+        # over again, in which case no other threads waiting on the
+        # "dhcp-agent" lock would make any progress.
+        eventlet.greenthread.sleep(0)
 
-    @utils.synchronized('dhcp-agent')
+    @_sync_lock
     def sync_state(self, networks=None):
         """Sync the local DHCP state with Neutron. If no networks are passed,
         or 'None' is one of the networks, sync all of the networks.
         """
         only_nets = set([] if (not networks or None in networks) else networks)
-        LOG.info(_LI('Synchronizing state'))
+        LOG.info('Synchronizing state')
         pool = eventlet.GreenPool(self.conf.num_sync_threads)
         known_network_ids = set(self.cache.get_network_ids())
 
         try:
-            active_networks = self.plugin_rpc.get_active_networks_info()
-            LOG.info(_LI('All active networks have been fetched through RPC.'))
+            active_networks = self.plugin_rpc.get_active_networks_info(
+                enable_dhcp_filter=False)
+            LOG.info('All active networks have been fetched through RPC.')
             active_network_ids = set(network.id for network in active_networks)
             for deleted_id in known_network_ids - active_network_ids:
                 try:
                     self.disable_dhcp_helper(deleted_id)
                 except Exception as e:
                     self.schedule_resync(e, deleted_id)
-                    LOG.exception(_LE('Unable to sync network state on '
-                                      'deleted network %s'), deleted_id)
+                    LOG.exception('Unable to sync network state on '
+                                  'deleted network %s', deleted_id)
 
             for network in active_networks:
                 if (not only_nets or  # specifically resync all
@@ -176,7 +227,7 @@ class DhcpAgent(manager.Manager):
             # we notify all ports in case some were created while the agent
             # was down
             self.dhcp_ready_ports |= set(self.cache.get_port_ids(only_nets))
-            LOG.info(_LI('Synchronizing state complete'))
+            LOG.info('Synchronizing state complete')
 
         except Exception as e:
             if only_nets:
@@ -184,7 +235,7 @@ class DhcpAgent(manager.Manager):
                     self.schedule_resync(e, network_id)
             else:
                 self.schedule_resync(e)
-            LOG.exception(_LE('Unable to sync network state.'))
+            LOG.exception('Unable to sync network state.')
 
     def _dhcp_ready_ports_loop(self):
         """Notifies the server of any ports that had reservations setup."""
@@ -198,19 +249,12 @@ class DhcpAgent(manager.Manager):
                     self.plugin_rpc.dhcp_ready_on_ports(ports_to_send)
                     continue
                 except oslo_messaging.MessagingTimeout:
-                    LOG.error(_LE("Timeout notifying server of ports ready. "
-                                  "Retrying..."))
-                except Exception as e:
-                    if (isinstance(e, oslo_messaging.RemoteError)
-                            and e.exc_type == 'NoSuchMethod'):
-                        LOG.info(_LI("Server does not support port ready "
-                                     "notifications. Waiting for 5 minutes "
-                                     "before retrying."))
-                        eventlet.sleep(300)
-                        continue
-                    LOG.exception(_LE("Failure notifying DHCP server of "
-                                      "ready DHCP ports. Will retry on next "
-                                      "iteration."))
+                    LOG.error("Timeout notifying server of ports ready. "
+                              "Retrying...")
+                except Exception:
+                    LOG.exception("Failure notifying DHCP server of "
+                                  "ready DHCP ports. Will retry on next "
+                                  "iteration.")
                 self.dhcp_ready_ports |= ports_to_send
 
     def start_ready_ports_loop(self):
@@ -219,9 +263,20 @@ class DhcpAgent(manager.Manager):
 
     @utils.exception_logger()
     def _periodic_resync_helper(self):
-        """Resync the dhcp state at the configured interval."""
+        """Resync the dhcp state at the configured interval and throttle."""
         while True:
-            eventlet.sleep(self.conf.resync_interval)
+            # threading.Event.wait blocks until the internal flag is true. It
+            # returns the internal flag on exit, so it will always return True
+            # except if a timeout is given and the operation times out.
+            if self._periodic_resync_event.wait(self.conf.resync_interval):
+                LOG.debug("Resync event has been scheduled")
+                clear_periodic_resync_event = self._periodic_resync_event.clear
+                # configure throttler for clear_periodic_resync_event to
+                # introduce delays between resync state events.
+                throttled_clear_periodic_resync_event = utils.throttler(
+                    self.conf.resync_throttle)(clear_periodic_resync_event)
+                throttled_clear_periodic_resync_event()
+
             if self.needs_resync_reasons:
                 # be careful to avoid a race with additions to list
                 # from other threads
@@ -246,7 +301,7 @@ class DhcpAgent(manager.Manager):
             return network
         except Exception as e:
             self.schedule_resync(e, network_id)
-            LOG.exception(_LE('Network %s info call failed.'), network_id)
+            LOG.exception('Network %s info call failed.', network_id)
 
     def enable_dhcp_helper(self, network_id):
         """Enable DHCP for a network that meets enabling criteria."""
@@ -258,12 +313,12 @@ class DhcpAgent(manager.Manager):
     def safe_configure_dhcp_for_network(self, network):
         try:
             network_id = network.get('id')
-            LOG.info(_LI('Starting network %s dhcp configuration'), network_id)
+            LOG.info('Starting network %s dhcp configuration', network_id)
             self.configure_dhcp_for_network(network)
-            LOG.info(_LI('Finished network %s dhcp configuration'), network_id)
+            LOG.info('Finished network %s dhcp configuration', network_id)
         except (exceptions.NetworkNotFound, RuntimeError):
-            LOG.warning(_LW('Network %s may have been deleted and '
-                            'its resources may have already been disposed.'),
+            LOG.warning('Network %s may have been deleted and '
+                        'its resources may have already been disposed.',
                         network.id)
 
     def configure_dhcp_for_network(self, network):
@@ -281,19 +336,25 @@ class DhcpAgent(manager.Manager):
                     self.dhcp_ready_ports |= {p.id for p in network.ports}
                 break
 
+        self._resize_process_pool()
+
     def disable_dhcp_helper(self, network_id):
         """Disable DHCP for a network known to the agent."""
         network = self.cache.get_network_by_id(network_id)
         if network:
-            if self.conf.enable_isolated_metadata:
-                # NOTE(jschwarz): In the case where a network is deleted, all
-                # the subnets and ports are deleted before this function is
-                # called, so checking if 'should_enable_metadata' is True
-                # for any subnet is false logic here.
-                self.disable_isolated_metadata_proxy(network)
+            # NOTE(yamahata): Kill the metadata proxy process
+            # unconditionally, as in the case where a network
+            # is deleted, all the subnets and ports are deleted
+            # before this function is called, so determining if
+            # the proxy should be terminated is error prone.
+            # destroy_monitored_metadata_proxy() is a noop when
+            # there is no process running.
+            self.disable_isolated_metadata_proxy(network)
             if self.call_driver('disable', network):
                 self.cache.remove(network)
 
+        self._resize_process_pool()
+
     def refresh_dhcp_helper(self, network_id):
         """Refresh or disable DHCP for a network depending on the current state
         of the network.
@@ -310,10 +371,12 @@ class DhcpAgent(manager.Manager):
         if not any(s for s in network.subnets if s.enable_dhcp):
             self.disable_dhcp_helper(network.id)
             return
-        # NOTE(kevinbenton): we don't exclude dhcp disabled subnets because
-        # they still change the indexes used for tags
-        old_cidrs = [s.cidr for s in old_network.subnets]
-        new_cidrs = [s.cidr for s in network.subnets]
+        old_non_local_subnets = getattr(old_network, 'non_local_subnets', [])
+        new_non_local_subnets = getattr(network, 'non_local_subnets', [])
+        old_cidrs = [s.cidr for s in (old_network.subnets +
+                                      old_non_local_subnets) if s.enable_dhcp]
+        new_cidrs = [s.cidr for s in (network.subnets +
+                                      new_non_local_subnets) if s.enable_dhcp]
         if old_cidrs == new_cidrs:
             self.call_driver('reload_allocations', network)
             self.cache.put(network)
@@ -326,104 +389,253 @@ class DhcpAgent(manager.Manager):
         # Update the metadata proxy after the dhcp driver has been updated
         self.update_isolated_metadata_proxy(network)
 
-    @utils.synchronized('dhcp-agent')
     def network_create_end(self, context, payload):
         """Handle the network.create.end notification event."""
+        update = queue.ResourceUpdate(payload['network']['id'],
+                                      payload.get('priority',
+                                                  DEFAULT_PRIORITY),
+                                      action='_network_create',
+                                      resource=payload)
+        self._queue.add(update)
+
+    @_wait_if_syncing
+    def _network_create(self, payload):
         network_id = payload['network']['id']
         self.enable_dhcp_helper(network_id)
 
-    @utils.synchronized('dhcp-agent')
     def network_update_end(self, context, payload):
         """Handle the network.update.end notification event."""
+        update = queue.ResourceUpdate(payload['network']['id'],
+                                      payload.get('priority',
+                                                  DEFAULT_PRIORITY),
+                                      action='_network_update',
+                                      resource=payload)
+        self._queue.add(update)
+
+    @_wait_if_syncing
+    def _network_update(self, payload):
         network_id = payload['network']['id']
         if payload['network']['admin_state_up']:
             self.enable_dhcp_helper(network_id)
         else:
             self.disable_dhcp_helper(network_id)
 
-    @utils.synchronized('dhcp-agent')
     def network_delete_end(self, context, payload):
         """Handle the network.delete.end notification event."""
-        self.disable_dhcp_helper(payload['network_id'])
+        update = queue.ResourceUpdate(payload['network_id'],
+                                      payload.get('priority',
+                                                  DEFAULT_PRIORITY),
+                                      action='_network_delete',
+                                      resource=payload)
+        self._queue.add(update)
+
+    @_wait_if_syncing
+    def _network_delete(self, payload):
+        network_id = payload['network_id']
+        self.disable_dhcp_helper(network_id)
 
-    @utils.synchronized('dhcp-agent')
     def subnet_update_end(self, context, payload):
         """Handle the subnet.update.end notification event."""
+        update = queue.ResourceUpdate(payload['subnet']['network_id'],
+                                      payload.get('priority',
+                                                  DEFAULT_PRIORITY),
+                                      action='_subnet_update',
+                                      resource=payload)
+        self._queue.add(update)
+
+    @_wait_if_syncing
+    def _subnet_update(self, payload):
         network_id = payload['subnet']['network_id']
         self.refresh_dhcp_helper(network_id)
 
     # Use the update handler for the subnet create event.
     subnet_create_end = subnet_update_end
 
-    @utils.synchronized('dhcp-agent')
+    def _get_network_lock_id(self, payload):
+        """Determine which lock to hold when servicing an RPC event"""
+        # TODO(alegacy): in a future release this function can be removed and
+        # uses of it can be replaced with payload['network_id'].  It exists
+        # only to satisfy backwards compatibility between older servers and
+        # newer agents.  Once the 'network_id' attribute is guaranteed to be
+        # sent by the server on all *_delete_end events then it can be removed.
+        if 'network_id' in payload:
+            return payload['network_id']
+        elif 'subnet_id' in payload:
+            subnet_id = payload['subnet_id']
+            network = self.cache.get_network_by_subnet_id(subnet_id)
+            return network.id if network else None
+        elif 'port_id' in payload:
+            port_id = payload['port_id']
+            port = self.cache.get_port_by_id(port_id)
+            return port.network_id if port else None
+
     def subnet_delete_end(self, context, payload):
         """Handle the subnet.delete.end notification event."""
+        network_id = self._get_network_lock_id(payload)
+        if not network_id:
+            return
+        update = queue.ResourceUpdate(network_id,
+                                      payload.get('priority',
+                                                  DEFAULT_PRIORITY),
+                                      action='_subnet_delete',
+                                      resource=payload)
+        self._queue.add(update)
+
+    @_wait_if_syncing
+    def _subnet_delete(self, payload):
+        network_id = self._get_network_lock_id(payload)
+        if not network_id:
+            return
         subnet_id = payload['subnet_id']
         network = self.cache.get_network_by_subnet_id(subnet_id)
-        if network:
-            self.refresh_dhcp_helper(network.id)
+        if not network:
+            return
+        self.refresh_dhcp_helper(network.id)
+
+    @lockutils.synchronized('resize_greenpool')
+    def _resize_process_pool(self):
+        num_nets = len(self.cache.get_network_ids())
+        pool_size = max([DHCP_PROCESS_GREENLET_MIN,
+                         min([DHCP_PROCESS_GREENLET_MAX, num_nets])])
+        if pool_size == self._pool_size:
+            return
+        LOG.info("Resizing dhcp processing queue green pool size to: %d",
+                 pool_size)
+        self._pool.resize(pool_size)
+        self._pool_size = pool_size
+
+    def _process_loop(self):
+        LOG.debug("Starting _process_loop")
+
+        while True:
+            self._pool.spawn_n(self._process_resource_update)
+
+    def _process_resource_update(self):
+        for tmp, update in self._queue.each_update_to_next_resource():
+            method = getattr(self, update.action)
+            method(update.resource)
 
-    @utils.synchronized('dhcp-agent')
     def port_update_end(self, context, payload):
         """Handle the port.update.end notification event."""
         updated_port = dhcp.DictModel(payload['port'])
-        if self.cache.is_port_message_stale(payload['port']):
+        if self.cache.is_port_message_stale(updated_port):
+            LOG.debug("Discarding stale port update: %s", updated_port)
+            return
+        update = queue.ResourceUpdate(updated_port.network_id,
+                                      payload.get('priority',
+                                                  DEFAULT_PRIORITY),
+                                      action='_port_update',
+                                      resource=updated_port)
+        self._queue.add(update)
+
+    @_wait_if_syncing
+    def _port_update(self, updated_port):
+        if self.cache.is_port_message_stale(updated_port):
             LOG.debug("Discarding stale port update: %s", updated_port)
             return
         network = self.cache.get_network_by_id(updated_port.network_id)
-        if network:
-            LOG.info(_LI("Trigger reload_allocations for port %s"),
-                     updated_port)
-            driver_action = 'reload_allocations'
-            if self._is_port_on_this_agent(updated_port):
-                orig = self.cache.get_port_by_id(updated_port['id'])
-                # assume IP change if not in cache
-                orig = orig or {'fixed_ips': []}
-                old_ips = {i['ip_address'] for i in orig['fixed_ips'] or []}
-                new_ips = {i['ip_address'] for i in updated_port['fixed_ips']}
-                old_subs = {i['subnet_id'] for i in orig['fixed_ips'] or []}
-                new_subs = {i['subnet_id'] for i in updated_port['fixed_ips']}
-                if new_subs != old_subs:
-                    # subnets being serviced by port have changed, this could
-                    # indicate a subnet_delete is in progress. schedule a
-                    # resync rather than an immediate restart so we don't
-                    # attempt to re-allocate IPs at the same time the server
-                    # is deleting them.
-                    self.schedule_resync("Agent port was modified",
-                                         updated_port.network_id)
-                    return
-                elif old_ips != new_ips:
-                    LOG.debug("Agent IPs on network %s changed from %s to %s",
-                              network.id, old_ips, new_ips)
-                    driver_action = 'restart'
-            self.cache.put_port(updated_port)
-            self.call_driver(driver_action, network)
-            self.dhcp_ready_ports.add(updated_port.id)
+        if not network:
+            return
+        self.reload_allocations(updated_port, network)
+
+    def reload_allocations(self, port, network):
+        LOG.info("Trigger reload_allocations for port %s", port)
+        driver_action = 'reload_allocations'
+        if self._is_port_on_this_agent(port):
+            orig = self.cache.get_port_by_id(port['id'])
+            # assume IP change if not in cache
+            orig = orig or {'fixed_ips': []}
+            old_ips = {i['ip_address'] for i in orig['fixed_ips'] or []}
+            new_ips = {i['ip_address'] for i in port['fixed_ips']}
+            old_subs = {i['subnet_id'] for i in orig['fixed_ips'] or []}
+            new_subs = {i['subnet_id'] for i in port['fixed_ips']}
+            if new_subs != old_subs:
+                # subnets being serviced by port have changed, this could
+                # indicate a subnet_delete is in progress. schedule a
+                # resync rather than an immediate restart so we don't
+                # attempt to re-allocate IPs at the same time the server
+                # is deleting them.
+                self.schedule_resync("Agent port was modified",
+                                     port.network_id)
+                return
+            elif old_ips != new_ips:
+                LOG.debug("Agent IPs on network %s changed from %s to %s",
+                          network.id, old_ips, new_ips)
+                driver_action = 'restart'
+        self.cache.put_port(port)
+        self.call_driver(driver_action, network)
+        self.dhcp_ready_ports.add(port.id)
+        self.update_isolated_metadata_proxy(network)
 
     def _is_port_on_this_agent(self, port):
         thishost = utils.get_dhcp_agent_device_id(
             port['network_id'], self.conf.host)
         return port['device_id'] == thishost
 
-    # Use the update handler for the port create event.
-    port_create_end = port_update_end
+    def port_create_end(self, context, payload):
+        """Handle the port.create.end notification event."""
+        created_port = dhcp.DictModel(payload['port'])
+        update = queue.ResourceUpdate(created_port.network_id,
+                                      payload.get('priority',
+                                                  DEFAULT_PRIORITY),
+                                      action='_port_create',
+                                      resource=created_port)
+        self._queue.add(update)
+
+    @_wait_if_syncing
+    def _port_create(self, created_port):
+        network = self.cache.get_network_by_id(created_port.network_id)
+        if not network:
+            return
+        new_ips = {i['ip_address'] for i in created_port['fixed_ips']}
+        for port_cached in network.ports:
+            # if in the same network there are ports cached with the same
+            # ip address but different MAC address and/or different id,
+            # this indicate that the cache is out of sync
+            cached_ips = {i['ip_address']
+                          for i in port_cached['fixed_ips']}
+            if (new_ips.intersection(cached_ips) and
+                (created_port['id'] != port_cached['id'] or
+                 created_port['mac_address'] != port_cached['mac_address'])):
+                self.schedule_resync("Duplicate IP addresses found, "
+                                     "DHCP cache is out of sync",
+                                     created_port.network_id)
+                return
+        self.reload_allocations(created_port, network)
 
-    @utils.synchronized('dhcp-agent')
     def port_delete_end(self, context, payload):
         """Handle the port.delete.end notification event."""
-        port = self.cache.get_port_by_id(payload['port_id'])
-        self.cache.deleted_ports.add(payload['port_id'])
-        if port:
-            network = self.cache.get_network_by_id(port.network_id)
-            self.cache.remove_port(port)
-            if self._is_port_on_this_agent(port):
-                # the agent's port has been deleted. disable the service
-                # and add the network to the resync list to create
-                # (or acquire a reserved) port.
-                self.call_driver('disable', network)
-                self.schedule_resync("Agent port was deleted", port.network_id)
-            else:
-                self.call_driver('reload_allocations', network)
+        network_id = self._get_network_lock_id(payload)
+        if not network_id:
+            return
+        update = queue.ResourceUpdate(network_id,
+                                      payload.get('priority',
+                                                  DEFAULT_PRIORITY),
+                                      action='_port_delete',
+                                      resource=payload)
+        self._queue.add(update)
+
+    @_wait_if_syncing
+    def _port_delete(self, payload):
+        network_id = self._get_network_lock_id(payload)
+        if not network_id:
+            return
+        port_id = payload['port_id']
+        port = self.cache.get_port_by_id(port_id)
+        self.cache.deleted_ports.add(port_id)
+        if not port:
+            return
+        network = self.cache.get_network_by_id(port.network_id)
+        self.cache.remove_port(port)
+        if self._is_port_on_this_agent(port):
+            # the agent's port has been deleted. disable the service
+            # and add the network to the resync list to create
+            # (or acquire a reserved) port.
+            self.call_driver('disable', network)
+            self.schedule_resync("Agent port was deleted", port.network_id)
+        else:
+            self.call_driver('reload_allocations', network)
+            self.update_isolated_metadata_proxy(network)
 
     def update_isolated_metadata_proxy(self, network):
         """Spawn or kill metadata proxy.
@@ -454,30 +666,33 @@ class DhcpAgent(manager.Manager):
             if router_ports:
                 # Multiple router ports should not be allowed
                 if len(router_ports) > 1:
-                    LOG.warning(_LW("%(port_num)d router ports found on the "
-                                    "metadata access network. Only the port "
-                                    "%(port_id)s, for router %(router_id)s "
-                                    "will be considered"),
+                    LOG.warning("%(port_num)d router ports found on the "
+                                "metadata access network. Only the port "
+                                "%(port_id)s, for router %(router_id)s "
+                                "will be considered",
                                 {'port_num': len(router_ports),
                                  'port_id': router_ports[0].id,
                                  'router_id': router_ports[0].device_id})
-                kwargs = {'router_id': router_ports[0].device_id}
-                self._metadata_routers[network.id] = router_ports[0].device_id
+                all_subnets = self.dhcp_driver_cls._get_all_subnets(network)
+                if self.dhcp_driver_cls.has_metadata_subnet(all_subnets):
+                    kwargs = {'router_id': router_ports[0].device_id}
+                    self._metadata_routers[network.id] = (
+                        router_ports[0].device_id)
 
         metadata_driver.MetadataDriver.spawn_monitored_metadata_proxy(
             self._process_monitor, network.namespace, dhcp.METADATA_PORT,
-            self.conf, **kwargs)
+            self.conf, bind_address=dhcp.METADATA_DEFAULT_IP, **kwargs)
 
     def disable_isolated_metadata_proxy(self, network):
         if (self.conf.enable_metadata_network and
-            network.id in self._metadata_routers):
+                network.id in self._metadata_routers):
             uuid = self._metadata_routers[network.id]
             is_router_id = True
         else:
             uuid = network.id
             is_router_id = False
         metadata_driver.MetadataDriver.destroy_monitored_metadata_proxy(
-            self._process_monitor, uuid, self.conf)
+            self._process_monitor, uuid, self.conf, network.namespace)
         if is_router_id:
             del self._metadata_routers[network.id]
 
@@ -488,7 +703,8 @@ class DhcpPluginApi(object):
     This class implements the client side of an rpc interface.  The server side
     of this interface can be found in
     neutron.api.rpc.handlers.dhcp_rpc.DhcpRpcCallback.  For more information
-    about changing rpc interfaces, see doc/source/devref/rpc_api.rst.
+    about changing rpc interfaces, see
+    doc/source/contributor/internals/rpc_api.rst.
 
     API version history:
         1.0 - Initial version.
@@ -515,11 +731,11 @@ class DhcpPluginApi(object):
         # can be independently tracked server side.
         return context.get_admin_context_without_session()
 
-    def get_active_networks_info(self):
+    def get_active_networks_info(self, **kwargs):
         """Make a remote process call to retrieve all network info."""
         cctxt = self.client.prepare(version='1.1')
         networks = cctxt.call(self.context, 'get_active_networks_info',
-                              host=self.host)
+                              host=self.host, **kwargs)
         return [dhcp.NetModel(n) for n in networks]
 
     def get_network_info(self, network_id):
@@ -600,7 +816,8 @@ class NetworkCache(object):
 
         self.cache[network.id] = network
 
-        for subnet in network.subnets:
+        non_local_subnets = getattr(network, 'non_local_subnets', [])
+        for subnet in (network.subnets + non_local_subnets):
             self.subnet_lookup[subnet.id] = network.id
 
         for port in network.ports:
@@ -609,7 +826,8 @@ class NetworkCache(object):
     def remove(self, network):
         del self.cache[network.id]
 
-        for subnet in network.subnets:
+        non_local_subnets = getattr(network, 'non_local_subnets', [])
+        for subnet in (network.subnets + non_local_subnets):
             del self.subnet_lookup[subnet.id]
 
         for port in network.ports:
@@ -649,7 +867,9 @@ class NetworkCache(object):
         num_ports = 0
         for net_id in net_ids:
             network = self.get_network_by_id(net_id)
+            non_local_subnets = getattr(network, 'non_local_subnets', [])
             num_subnets += len(network.subnets)
+            num_subnets += len(non_local_subnets)
             num_ports += len(network.ports)
         return {'networks': num_nets,
                 'subnets': num_subnets,
@@ -660,13 +880,13 @@ class DhcpAgentWithStateReport(DhcpAgent):
     def __init__(self, host=None, conf=None):
         super(DhcpAgentWithStateReport, self).__init__(host=host, conf=conf)
         self.state_rpc = agent_rpc.PluginReportStateAPI(topics.REPORTS)
+        self.failed_report_state = False
         self.agent_state = {
             'binary': 'neutron-dhcp-agent',
             'host': host,
             'availability_zone': self.conf.AGENT.availability_zone,
             'topic': topics.DHCP_AGENT,
             'configurations': {
-                'notifies_port_ready': True,
                 'dhcp_driver': self.conf.dhcp_driver,
                 'dhcp_lease_duration': self.conf.dhcp_lease_duration,
                 'log_agent_heartbeats': self.conf.AGENT.log_agent_heartbeats},
@@ -685,20 +905,24 @@ class DhcpAgentWithStateReport(DhcpAgent):
             ctx = context.get_admin_context_without_session()
             agent_status = self.state_rpc.report_state(
                 ctx, self.agent_state, True)
-            if agent_status == n_const.AGENT_REVIVED:
-                LOG.info(_LI("Agent has just been revived. "
-                             "Scheduling full sync"))
+            if agent_status == agent_consts.AGENT_REVIVED:
+                LOG.info("Agent has just been revived. "
+                         "Scheduling full sync")
                 self.schedule_resync("Agent has just been revived")
         except AttributeError:
             # This means the server does not support report_state
-            LOG.warning(_LW("Neutron server does not support state report. "
-                            "State report for this agent will be disabled."))
+            LOG.warning("Neutron server does not support state report. "
+                        "State report for this agent will be disabled.")
             self.heartbeat.stop()
             self.run()
             return
         except Exception:
-            LOG.exception(_LE("Failed reporting state!"))
+            self.failed_report_state = True
+            LOG.exception("Failed reporting state!")
             return
+        if self.failed_report_state:
+            self.failed_report_state = False
+            LOG.info("Successfully reported state after a previous failure.")
         if self.agent_state.pop('start_flag', None):
             self.run()
 
@@ -706,7 +930,7 @@ class DhcpAgentWithStateReport(DhcpAgent):
         """Handle the agent_updated notification event."""
         self.schedule_resync(_("Agent updated: %(payload)s") %
                              {"payload": payload})
-        LOG.info(_LI("agent_updated by server side %s!"), payload)
+        LOG.info("agent_updated by server side %s!", payload)
 
     def after_start(self):
-        LOG.info(_LI("DHCP agent started"))
+        LOG.info("DHCP agent started")
