@@ -15,24 +15,27 @@
 from neutron_lib import constants as lib_constants
 from oslo_log import log as logging
 
-from neutron._i18n import _LE
 from neutron.agent.l3 import dvr_local_router
 from neutron.agent.l3 import dvr_snat_ns
 from neutron.agent.l3 import router_info as router
 from neutron.agent.linux import ip_lib
 from neutron.agent.linux import iptables_manager
+from neutron.common import utils as common_utils
 
 LOG = logging.getLogger(__name__)
 
 
 class DvrEdgeRouter(dvr_local_router.DvrLocalRouter):
 
-    def __init__(self, agent, host, *args, **kwargs):
-        super(DvrEdgeRouter, self).__init__(agent, host, *args, **kwargs)
+    def __init__(self, host, *args, **kwargs):
+        super(DvrEdgeRouter, self).__init__(host, *args, **kwargs)
         self.snat_namespace = dvr_snat_ns.SnatNamespace(
             self.router_id, self.agent_conf, self.driver, self.use_ipv6)
         self.snat_iptables_manager = None
 
+    def get_gw_ns_name(self):
+        return self.snat_namespace.name
+
     def external_gateway_added(self, ex_gw_port, interface_name):
         super(DvrEdgeRouter, self).external_gateway_added(
             ex_gw_port, interface_name)
@@ -52,6 +55,15 @@ class DvrEdgeRouter(dvr_local_router.DvrLocalRouter):
                       "current dvr_snat host.", self.snat_namespace.name)
             self.external_gateway_removed(ex_gw_port, interface_name)
 
+    def _list_centralized_floating_ip_cidrs(self):
+        # Compute a list of addresses this gw is supposed to have.
+        # This avoids unnecessarily removing those addresses and
+        # causing a momentarily network outage.
+        floating_ips = self.get_floating_ips()
+        return [common_utils.ip_to_cidr(ip['floating_ip_address'])
+                for ip in floating_ips
+                if ip.get(lib_constants.DVR_SNAT_BOUND)]
+
     def external_gateway_updated(self, ex_gw_port, interface_name):
         if not self._is_this_snat_host():
             # no centralized SNAT gateway for this node/agent
@@ -68,10 +80,11 @@ class DvrEdgeRouter(dvr_local_router.DvrLocalRouter):
             # newly created gateway
             return self.external_gateway_added(ex_gw_port, interface_name)
         else:
+            preserve_ips = self._list_centralized_floating_ip_cidrs()
             self._external_gateway_added(ex_gw_port,
-                                        interface_name,
-                                        self.snat_namespace.name,
-                                        preserve_ips=[])
+                                         interface_name,
+                                         self.snat_namespace.name,
+                                         preserve_ips)
 
     def _external_gateway_removed(self, ex_gw_port, interface_name):
         super(DvrEdgeRouter, self).external_gateway_removed(ex_gw_port,
@@ -82,7 +95,6 @@ class DvrEdgeRouter(dvr_local_router.DvrLocalRouter):
             return
 
         self.driver.unplug(interface_name,
-                           bridge=self.agent_conf.external_network_bridge,
                            namespace=self.snat_namespace.name,
                            prefix=router.EXTERNAL_DEV_PREFIX)
 
@@ -112,7 +124,7 @@ class DvrEdgeRouter(dvr_local_router.DvrLocalRouter):
             sn_port['fixed_ips'],
             sn_port['mac_address'],
             interface_name,
-            dvr_snat_ns.SNAT_INT_DEV_PREFIX,
+            lib_constants.SNAT_INT_DEV_PREFIX,
             mtu=sn_port.get('mtu'))
 
     def _dvr_internal_network_removed(self, port):
@@ -130,7 +142,7 @@ class DvrEdgeRouter(dvr_local_router.DvrLocalRouter):
 
         snat_interface = self._get_snat_int_device_name(sn_port['id'])
         ns_name = self.snat_namespace.name
-        prefix = dvr_snat_ns.SNAT_INT_DEV_PREFIX
+        prefix = lib_constants.SNAT_INT_DEV_PREFIX
         if ip_lib.device_exists(snat_interface, namespace=ns_name):
             self.driver.unplug(snat_interface, namespace=ns_name,
                                prefix=prefix)
@@ -141,15 +153,13 @@ class DvrEdgeRouter(dvr_local_router.DvrLocalRouter):
             self.snat_namespace.name, port['network_id'],
             port['id'], port['fixed_ips'],
             port['mac_address'], interface_name,
-            dvr_snat_ns.SNAT_INT_DEV_PREFIX,
+            lib_constants.SNAT_INT_DEV_PREFIX,
             mtu=port.get('mtu'))
 
     def _create_dvr_gateway(self, ex_gw_port, gw_interface_name):
-        """Create SNAT namespace."""
         snat_ns = self._create_snat_namespace()
         # connect snat_ports to br_int from SNAT namespace
         for port in self.get_snat_interfaces():
-            # create interface_name
             self._plug_snat_port(port)
         self._external_gateway_added(ex_gw_port, gw_interface_name,
                                      snat_ns.name, preserve_ips=[])
@@ -160,6 +170,7 @@ class DvrEdgeRouter(dvr_local_router.DvrLocalRouter):
         self._initialize_address_scope_iptables(self.snat_iptables_manager)
 
     def _create_snat_namespace(self):
+        """Create SNAT namespace."""
         # TODO(mlavalle): in the near future, this method should contain the
         # code in the L3 agent that creates a gateway for a dvr. The first step
         # is to move the creation of the snat namespace here
@@ -167,7 +178,7 @@ class DvrEdgeRouter(dvr_local_router.DvrLocalRouter):
         return self.snat_namespace
 
     def _get_snat_int_device_name(self, port_id):
-        long_name = dvr_snat_ns.SNAT_INT_DEV_PREFIX + port_id
+        long_name = lib_constants.SNAT_INT_DEV_PREFIX + port_id
         return long_name[:self.driver.DEV_NAME_LEN]
 
     def _is_this_snat_host(self):
@@ -193,8 +204,11 @@ class DvrEdgeRouter(dvr_local_router.DvrLocalRouter):
         with self.snat_iptables_manager.defer_apply():
             self._empty_snat_chains(self.snat_iptables_manager)
 
-            # NOTE: DVR adds the jump to float snat via super class,
-            # but that is in the router namespace and not snat.
+            # NOTE: float-snat should be added for the
+            # centralized floating-ips supported by the
+            # snat namespace.
+            self.snat_iptables_manager.ipv4['nat'].add_rule(
+                'snat', '-j $float-snat')
 
             self._add_snat_rules(ex_gw_port, self.snat_iptables_manager,
                                  interface_name)
@@ -208,12 +222,12 @@ class DvrEdgeRouter(dvr_local_router.DvrLocalRouter):
                 super(DvrEdgeRouter, self)._update_routing_table(
                     operation, route, namespace=ns_name)
             else:
-                LOG.error(_LE("The SNAT namespace %s does not exist for "
-                              "the router."), ns_name)
+                LOG.error("The SNAT namespace %s does not exist for "
+                          "the router.", ns_name)
         super(DvrEdgeRouter, self).update_routing_table(operation, route)
 
-    def delete(self, agent):
-        super(DvrEdgeRouter, self).delete(agent)
+    def delete(self):
+        super(DvrEdgeRouter, self).delete()
         if self.snat_namespace.exists():
             self.snat_namespace.delete()
 
@@ -243,3 +257,120 @@ class DvrEdgeRouter(dvr_local_router.DvrLocalRouter):
         with self.snat_iptables_manager.defer_apply():
             self._add_address_scope_mark(
                 self.snat_iptables_manager, ports_scopemark)
+
+    def _delete_stale_external_devices(self, interface_name):
+        if not self.snat_namespace.exists():
+            return
+
+        ns_ip = ip_lib.IPWrapper(namespace=self.snat_namespace.name)
+        for d in ns_ip.get_devices():
+            if (d.name.startswith(router.EXTERNAL_DEV_PREFIX) and
+                    d.name != interface_name):
+                LOG.debug('Deleting stale external router device: %s', d.name)
+                self.driver.unplug(
+                    d.name,
+                    namespace=self.snat_namespace.name,
+                    prefix=router.EXTERNAL_DEV_PREFIX)
+
+    def get_snat_external_device_interface_name(self, ex_gw_port):
+        long_name = router.EXTERNAL_DEV_PREFIX + ex_gw_port['id']
+        return long_name[:self.driver.DEV_NAME_LEN]
+
+    def get_centralized_fip_cidr_set(self):
+        """Returns the fip_cidr set for centralized floatingips."""
+        ex_gw_port = self.get_ex_gw_port()
+        # Don't look for centralized FIP cidrs if gw_port not exists or
+        # this is not snat host
+        if (not ex_gw_port or not self._is_this_snat_host() or
+                not self.snat_namespace.exists()):
+            return set()
+        interface_name = self.get_snat_external_device_interface_name(
+                ex_gw_port)
+        return set([addr['cidr'] for addr in ip_lib.get_devices_with_ip(
+                                                 self.snat_namespace.name,
+                                                 name=interface_name)])
+
+    def get_router_cidrs(self, device):
+        """Over-ride the get_router_cidrs function to return the list.
+
+        This function is overridden to provide the complete list of
+        floating_ip cidrs that the router hosts.
+        This includes the centralized floatingip cidr list and the
+        regular floatingip cidr list that are bound to fip namespace.
+        """
+        fip_cidrs = super(DvrEdgeRouter, self).get_router_cidrs(device)
+        centralized_cidrs = self.get_centralized_fip_cidr_set()
+        return fip_cidrs | centralized_cidrs
+
+    def remove_centralized_floatingip(self, fip_cidr):
+        """Function to handle the centralized Floatingip remove."""
+        if not self.get_ex_gw_port():
+            return
+        if not self._is_this_snat_host():
+            return
+        interface_name = self.get_snat_external_device_interface_name(
+            self.get_ex_gw_port())
+        device = ip_lib.IPDevice(
+            interface_name, namespace=self.snat_namespace.name)
+        device.delete_addr_and_conntrack_state(fip_cidr)
+        self.process_floating_ip_nat_rules_for_centralized_floatingip()
+
+    def add_centralized_floatingip(self, fip, fip_cidr):
+        """Function to handle the centralized Floatingip addition."""
+        if not self.get_ex_gw_port():
+            return
+        if not self._is_this_snat_host():
+            return
+        interface_name = self.get_snat_external_device_interface_name(
+            self.get_ex_gw_port())
+        try:
+            ip_lib.add_ip_address(fip_cidr, interface_name,
+                                  namespace=self.snat_namespace.name)
+        except ip_lib.IpAddressAlreadyExists:
+            pass
+        except RuntimeError:
+            LOG.warning("Unable to configure IP address for centralized "
+                        "floating IP: %s", fip['id'])
+            return lib_constants.FLOATINGIP_STATUS_ERROR
+        self.process_floating_ip_nat_rules_for_centralized_floatingip()
+        # Send a GARP message on the external interface for the
+        # centralized floatingip configured.
+        ip_lib.send_ip_addr_adv_notif(self.snat_namespace.name,
+                                      interface_name,
+                                      fip['floating_ip_address'])
+        return lib_constants.FLOATINGIP_STATUS_ACTIVE
+
+    def _centralized_floating_forward_rules(self, floating_ip, fixed_ip):
+        to_source = '-s %s/32 -j SNAT --to-source %s' % (fixed_ip, floating_ip)
+        if self.snat_iptables_manager.random_fully:
+            to_source += ' --random-fully'
+        return [('PREROUTING', '-d %s/32 -j DNAT --to-destination %s' %
+                 (floating_ip, fixed_ip)),
+                ('OUTPUT', '-d %s/32 -j DNAT --to-destination %s' %
+                 (floating_ip, fixed_ip)),
+                ('float-snat', to_source)]
+
+    def _set_floating_ip_nat_rules_for_centralized_floatingip(self, fip):
+        if fip.get(lib_constants.DVR_SNAT_BOUND):
+            fixed = fip['fixed_ip_address']
+            fip_ip = fip['floating_ip_address']
+            for chain, rule in self._centralized_floating_forward_rules(
+                fip_ip, fixed):
+                self.snat_iptables_manager.ipv4['nat'].add_rule(
+                    chain, rule, tag='floating_ip')
+
+    def process_floating_ip_nat_rules_for_centralized_floatingip(self):
+        self.snat_iptables_manager.ipv4['nat'].clear_rules_by_tag(
+            'floating_ip')
+        floating_ips = self.get_floating_ips()
+        for fip in floating_ips:
+            self._set_floating_ip_nat_rules_for_centralized_floatingip(fip)
+        self.snat_iptables_manager.apply()
+
+    def process_floating_ip_nat_rules(self):
+        if self._is_this_snat_host():
+            self.process_floating_ip_nat_rules_for_centralized_floatingip()
+
+        # Cover mixed dvr_snat and compute node, aka a dvr_snat node has both
+        # centralized and distributed floating IPs.
+        super(DvrEdgeRouter, self).process_floating_ip_nat_rules()
