@@ -13,17 +13,19 @@
 #    under the License.
 
 import collections
+
 import netaddr
 from neutron_lib import constants as lib_constants
+from neutron_lib.exceptions import l3 as l3_exc
+from neutron_lib.utils import helpers
 from oslo_log import log as logging
 
-from neutron._i18n import _, _LE, _LW
+from neutron._i18n import _
 from neutron.agent.l3 import namespaces
 from neutron.agent.linux import ip_lib
 from neutron.agent.linux import iptables_manager
 from neutron.agent.linux import ra
 from neutron.common import constants as n_const
-from neutron.common import exceptions as n_exc
 from neutron.common import ipv6_utils
 from neutron.common import utils as common_utils
 from neutron.ipam import utils as ipam_utils
@@ -42,21 +44,25 @@ DEFAULT_ADDRESS_SCOPE = "noscope"
 class RouterInfo(object):
 
     def __init__(self,
+                 agent,
                  router_id,
                  router,
                  agent_conf,
                  interface_driver,
                  use_ipv6=False):
+        self.agent = agent
         self.router_id = router_id
+        self.agent_conf = agent_conf
         self.ex_gw_port = None
         self._snat_enabled = None
         self.fip_map = {}
         self.internal_ports = []
+        self.pd_subnets = {}
         self.floating_ips = set()
         # Invoke the setter for establishing initial SNAT action
         self.router = router
         self.use_ipv6 = use_ipv6
-        ns = namespaces.RouterNamespace(
+        ns = self.create_router_namespace_object(
             router_id, agent_conf, interface_driver, use_ipv6)
         self.router_namespace = ns
         self.ns_name = ns.name
@@ -68,11 +74,15 @@ class RouterInfo(object):
             use_ipv6=use_ipv6,
             namespace=self.ns_name)
         self.initialize_address_scope_iptables()
+        self.initialize_metadata_iptables()
         self.routes = []
-        self.agent_conf = agent_conf
         self.driver = interface_driver
+        self.process_monitor = None
         # radvd is a neutron.agent.linux.ra.DaemonMonitor
         self.radvd = None
+        self.centralized_port_forwarding_fip_set = set()
+        self.fip_managed_by_port_forwardings = None
+        self.qos_gateway_ips = set()
 
     def initialize(self, process_monitor):
         """Initialize the router on the system.
@@ -93,6 +103,11 @@ class RouterInfo(object):
 
         self.router_namespace.create()
 
+    def create_router_namespace_object(
+            self, router_id, agent_conf, iface_driver, use_ipv6):
+        return namespaces.RouterNamespace(
+            router_id, agent_conf, iface_driver, use_ipv6)
+
     @property
     def router(self):
         return self._router
@@ -105,6 +120,9 @@ class RouterInfo(object):
         # enable_snat by default if it wasn't specified by plugin
         self._snat_enabled = self._router.get('enable_snat', True)
 
+    def is_router_master(self):
+        return True
+
     def get_internal_device_name(self, port_id):
         return (INTERNAL_DEV_PREFIX + port_id)[:self.driver.DEV_NAME_LEN]
 
@@ -114,6 +132,9 @@ class RouterInfo(object):
     def get_external_device_interface_name(self, ex_gw_port):
         return self.get_external_device_name(ex_gw_port['id'])
 
+    def get_gw_ns_name(self):
+        return self.ns_name
+
     def _update_routing_table(self, operation, route, namespace):
         cmd = ['ip', 'route', operation, 'to', route['destination'],
                'via', route['nexthop']]
@@ -124,15 +145,15 @@ class RouterInfo(object):
         self._update_routing_table(operation, route, self.ns_name)
 
     def routes_updated(self, old_routes, new_routes):
-        adds, removes = common_utils.diff_list_of_dict(old_routes,
-                                                       new_routes)
+        adds, removes = helpers.diff_list_of_dict(old_routes,
+                                                  new_routes)
         for route in adds:
             LOG.debug("Added route entry is '%s'", route)
             # remove replaced route from deleted route
             for del_route in removes:
                 if route['destination'] == del_route['destination']:
                     removes.remove(del_route)
-            #replace success even if there is no existing route
+            # replace success even if there is no existing route
             self.update_routing_table('replace', route)
         for route in removes:
             LOG.debug("Removed route entry is '%s'", route)
@@ -145,13 +166,21 @@ class RouterInfo(object):
         """Filter Floating IPs to be hosted on this agent."""
         return self.router.get(lib_constants.FLOATINGIP_KEY, [])
 
-    def floating_forward_rules(self, floating_ip, fixed_ip):
+    def get_port_forwarding_fips(self):
+        """Get router port forwarding floating IPs."""
+        return self.router.get('_pf_floatingips', [])
+
+    def floating_forward_rules(self, fip):
+        fixed_ip = fip['fixed_ip_address']
+        floating_ip = fip['floating_ip_address']
+        to_source = '-s %s/32 -j SNAT --to-source %s' % (fixed_ip, floating_ip)
+        if self.iptables_manager.random_fully:
+            to_source += ' --random-fully'
         return [('PREROUTING', '-d %s/32 -j DNAT --to-destination %s' %
                  (floating_ip, fixed_ip)),
                 ('OUTPUT', '-d %s/32 -j DNAT --to-destination %s' %
                  (floating_ip, fixed_ip)),
-                ('float-snat', '-s %s/32 -j SNAT --to-source %s' %
-                 (fixed_ip, floating_ip))]
+                ('float-snat', to_source)]
 
     def floating_mangle_rules(self, floating_ip, fixed_ip, internal_mark):
         mark_traffic_to_floating_ip = (
@@ -199,14 +228,25 @@ class RouterInfo(object):
         # Loop once to ensure that floating ips are configured.
         for fip in floating_ips:
             # Rebuild iptables rules for the floating ip.
-            fixed = fip['fixed_ip_address']
-            fip_ip = fip['floating_ip_address']
-            for chain, rule in self.floating_forward_rules(fip_ip, fixed):
+            for chain, rule in self.floating_forward_rules(fip):
                 self.iptables_manager.ipv4['nat'].add_rule(chain, rule,
                                                            tag='floating_ip')
 
         self.iptables_manager.apply()
 
+    def _process_pd_iptables_rules(self, prefix, subnet_id):
+        """Configure iptables rules for prefix delegated subnets"""
+        ext_scope = self._get_external_address_scope()
+        ext_scope_mark = self.get_address_scope_mark_mask(ext_scope)
+        ex_gw_device = self.get_external_device_name(
+            self.get_ex_gw_port()['id'])
+        scope_rule = self.address_scope_mangle_rule(ex_gw_device,
+                                                    ext_scope_mark)
+        self.iptables_manager.ipv6['mangle'].add_rule(
+            'scope',
+            '-d %s ' % prefix + scope_rule,
+            tag=('prefix_delegation_%s' % subnet_id))
+
     def process_floating_ip_address_scope_rules(self):
         """Configure address scope related iptables rules for the router's
          floating IPs.
@@ -257,7 +297,7 @@ class RouterInfo(object):
             # TODO(salv-orlando): Less broad catching
             msg = _('L3 agent failure to setup NAT for floating IPs')
             LOG.exception(msg)
-            raise n_exc.FloatingIpSetupException(msg)
+            raise l3_exc.FloatingIpSetupException(msg)
 
     def _add_fip_addr_to_device(self, fip, device):
         """Configures the floating ip address on the device.
@@ -269,12 +309,18 @@ class RouterInfo(object):
         except RuntimeError:
             # any exception occurred here should cause the floating IP
             # to be set in error state
-            LOG.warning(_LW("Unable to configure IP address for "
-                            "floating IP: %s"), fip['id'])
+            LOG.warning("Unable to configure IP address for "
+                        "floating IP: %s", fip['id'])
 
     def add_floating_ip(self, fip, interface_name, device):
         raise NotImplementedError()
 
+    def migrate_centralized_floating_ip(self, fip, interface_name, device):
+        """Implements centralized->distributed floating IP migration.
+        Overridden in dvr_local_router.py
+        """
+        return FLOATINGIP_STATUS_NOCHANGE
+
     def gateway_redirect_cleanup(self, rtr_interface):
         pass
 
@@ -290,6 +336,9 @@ class RouterInfo(object):
     def get_router_cidrs(self, device):
         return set([addr['cidr'] for addr in device.addr.list()])
 
+    def get_centralized_fip_cidr_set(self):
+        return set()
+
     def process_floating_ip_addresses(self, interface_name):
         """Configure IP addresses on router's external gateway interface.
 
@@ -307,7 +356,7 @@ class RouterInfo(object):
         existing_cidrs = self.get_router_cidrs(device)
         new_cidrs = set()
         gw_cidrs = self._get_gw_ips_cidr()
-
+        centralized_fip_cidrs = self.get_centralized_fip_cidr_set()
         floating_ips = self.get_floating_ips()
         # Loop once to ensure that floating ips are configured.
         for fip in floating_ips:
@@ -315,6 +364,7 @@ class RouterInfo(object):
             ip_cidr = common_utils.ip_to_cidr(fip_ip)
             new_cidrs.add(ip_cidr)
             fip_statuses[fip['id']] = lib_constants.FLOATINGIP_STATUS_ACTIVE
+
             if ip_cidr not in existing_cidrs:
                 fip_statuses[fip['id']] = self.add_floating_ip(
                     fip, interface_name, device)
@@ -328,12 +378,20 @@ class RouterInfo(object):
                           {'old': self.fip_map[fip_ip],
                            'new': fip['fixed_ip_address']})
                 fip_statuses[fip['id']] = self.move_floating_ip(fip)
+            elif (ip_cidr in centralized_fip_cidrs and
+                  fip.get('host') == self.host):
+                LOG.debug("Floating IP is migrating from centralized "
+                          "to distributed: %s", fip)
+                fip_statuses[fip['id']] = self.migrate_centralized_floating_ip(
+                    fip, interface_name, device)
             elif fip_statuses[fip['id']] == fip['status']:
                 # mark the status as not changed. we can't remove it because
                 # that's how the caller determines that it was removed
                 fip_statuses[fip['id']] = FLOATINGIP_STATUS_NOCHANGE
         fips_to_remove = (
-            ip_cidr for ip_cidr in existing_cidrs - new_cidrs - gw_cidrs
+            ip_cidr
+            for ip_cidr in (existing_cidrs - new_cidrs - gw_cidrs -
+                            self.centralized_port_forwarding_fip_set)
             if common_utils.is_cidr_host(ip_cidr))
         for ip_cidr in fips_to_remove:
             LOG.debug("Removing floating ip %s from interface %s in "
@@ -360,7 +418,7 @@ class RouterInfo(object):
             # TODO(salv-orlando): Less broad catching
             msg = _('L3 agent failure to setup floating IPs')
             LOG.exception(msg)
-            raise n_exc.FloatingIpSetupException(msg)
+            raise l3_exc.FloatingIpSetupException(msg)
 
     def put_fips_in_error_state(self):
         fip_statuses = {}
@@ -368,18 +426,18 @@ class RouterInfo(object):
             fip_statuses[fip['id']] = lib_constants.FLOATINGIP_STATUS_ERROR
         return fip_statuses
 
-    def delete(self, agent):
+    def delete(self):
         self.router['gw_port'] = None
         self.router[lib_constants.INTERFACE_KEY] = []
         self.router[lib_constants.FLOATINGIP_KEY] = []
-        self.process_delete(agent)
+        self.process_delete()
         self.disable_radvd()
         self.router_namespace.delete()
 
     def _internal_network_updated(self, port, subnet_id, prefix, old_prefix,
                                   updated_cidrs):
         interface_name = self.get_internal_device_name(port['id'])
-        if prefix != n_const.PROVISIONAL_IPV6_PD_PREFIX:
+        if prefix != lib_constants.PROVISIONAL_IPV6_PD_PREFIX:
             fixed_ips = port['fixed_ips']
             for fixed_ip in fixed_ips:
                 if fixed_ip['subnet_id'] == subnet_id:
@@ -408,8 +466,7 @@ class RouterInfo(object):
         for fixed_ip in fixed_ips:
             ip_lib.send_ip_addr_adv_notif(ns_name,
                                           interface_name,
-                                          fixed_ip['ip_address'],
-                                          self.agent_conf)
+                                          fixed_ip['ip_address'])
 
     def internal_network_added(self, port):
         network_id = port['network_id']
@@ -438,21 +495,38 @@ class RouterInfo(object):
 
     def _get_existing_devices(self):
         ip_wrapper = ip_lib.IPWrapper(namespace=self.ns_name)
-        ip_devs = ip_wrapper.get_devices(exclude_loopback=True)
+        ip_devs = ip_wrapper.get_devices()
         return [ip_dev.name for ip_dev in ip_devs]
 
+    def _update_internal_ports_cache(self, port):
+        # NOTE(slaweq): self.internal_ports is a list of port objects but
+        # when it is updated in _process_internal_ports() method,
+        # but it can be based only on indexes of elements in
+        # self.internal_ports as index of element to updated is unknown.
+        # It has to be done based on port_id and this method is doing exactly
+        # that.
+        for index, p in enumerate(self.internal_ports):
+            if p['id'] == port['id']:
+                self.internal_ports[index] = port
+                break
+        else:
+            self.internal_ports.append(port)
+
     @staticmethod
     def _get_updated_ports(existing_ports, current_ports):
-        updated_ports = dict()
+        updated_ports = []
         current_ports_dict = {p['id']: p for p in current_ports}
         for existing_port in existing_ports:
             current_port = current_ports_dict.get(existing_port['id'])
             if current_port:
-                if (sorted(existing_port['fixed_ips'],
-                           key=common_utils.safe_sort_key) !=
-                        sorted(current_port['fixed_ips'],
-                               key=common_utils.safe_sort_key)):
-                    updated_ports[current_port['id']] = current_port
+                fixed_ips_changed = (
+                    sorted(existing_port['fixed_ips'],
+                           key=helpers.safe_sort_key) !=
+                    sorted(current_port['fixed_ips'],
+                           key=helpers.safe_sort_key))
+                mtu_changed = existing_port['mtu'] != current_port['mtu']
+                if fixed_ips_changed or mtu_changed:
+                    updated_ports.append(current_port)
         return updated_ports
 
     @staticmethod
@@ -460,7 +534,8 @@ class RouterInfo(object):
         if 'subnets' in port:
             for subnet in port['subnets']:
                 if (netaddr.IPNetwork(subnet['cidr']).version == 6 and
-                    subnet['cidr'] != n_const.PROVISIONAL_IPV6_PD_PREFIX):
+                    subnet['cidr'] !=
+                        lib_constants.PROVISIONAL_IPV6_PD_PREFIX):
                     return True
 
     def enable_radvd(self, internal_ports=None):
@@ -470,11 +545,14 @@ class RouterInfo(object):
         self.radvd.enable(internal_ports)
 
     def disable_radvd(self):
-        LOG.debug('Terminating radvd daemon in router device: %s',
-                  self.router_id)
-        self.radvd.disable()
+        if self.radvd:
+            LOG.debug('Terminating radvd daemon in router device: %s',
+                      self.router_id)
+            self.radvd.disable()
 
-    def internal_network_updated(self, interface_name, ip_cidrs):
+    def internal_network_updated(self, interface_name, ip_cidrs, mtu):
+        self.driver.set_mtu(interface_name, mtu, namespace=self.ns_name,
+                            prefix=INTERNAL_DEV_PREFIX)
         self.driver.init_router_port(
             interface_name,
             ip_cidrs=ip_cidrs,
@@ -487,7 +565,7 @@ class RouterInfo(object):
         return '-o %s -m mark ! --mark %s -j DROP' % (
             device_name, mark_mask)
 
-    def _process_internal_ports(self, pd):
+    def _process_internal_ports(self):
         existing_port_ids = set(p['id'] for p in self.internal_ports)
 
         internal_ports = self.router.get(lib_constants.INTERFACE_KEY, [])
@@ -505,14 +583,18 @@ class RouterInfo(object):
         for p in new_ports:
             self.internal_network_added(p)
             LOG.debug("appending port %s to internal_ports cache", p)
-            self.internal_ports.append(p)
+            self._update_internal_ports_cache(p)
             enable_ra = enable_ra or self._port_has_ipv6_subnet(p)
             for subnet in p['subnets']:
                 if ipv6_utils.is_ipv6_pd_enabled(subnet):
                     interface_name = self.get_internal_device_name(p['id'])
-                    pd.enable_subnet(self.router_id, subnet['id'],
-                                     subnet['cidr'],
-                                     interface_name, p['mac_address'])
+                    self.agent.pd.enable_subnet(self.router_id, subnet['id'],
+                                                subnet['cidr'],
+                                                interface_name,
+                                                p['mac_address'])
+                    if (subnet['cidr'] !=
+                            lib_constants.PROVISIONAL_IPV6_PD_PREFIX):
+                        self.pd_subnets[subnet['id']] = subnet['cidr']
 
         for p in old_ports:
             self.internal_network_removed(p)
@@ -521,27 +603,27 @@ class RouterInfo(object):
             enable_ra = enable_ra or self._port_has_ipv6_subnet(p)
             for subnet in p['subnets']:
                 if ipv6_utils.is_ipv6_pd_enabled(subnet):
-                    pd.disable_subnet(self.router_id, subnet['id'])
+                    self.agent.pd.disable_subnet(self.router_id, subnet['id'])
+                    del self.pd_subnets[subnet['id']]
 
         updated_cidrs = []
-        if updated_ports:
-            for index, p in enumerate(internal_ports):
-                if not updated_ports.get(p['id']):
-                    continue
-                self.internal_ports[index] = updated_ports[p['id']]
-                interface_name = self.get_internal_device_name(p['id'])
-                ip_cidrs = common_utils.fixed_ip_cidrs(p['fixed_ips'])
-                LOG.debug("updating internal network for port %s", p)
-                updated_cidrs += ip_cidrs
-                self.internal_network_updated(interface_name, ip_cidrs)
-                enable_ra = enable_ra or self._port_has_ipv6_subnet(p)
+        for p in updated_ports:
+            self._update_internal_ports_cache(p)
+            interface_name = self.get_internal_device_name(p['id'])
+            ip_cidrs = common_utils.fixed_ip_cidrs(p['fixed_ips'])
+            LOG.debug("updating internal network for port %s", p)
+            updated_cidrs += ip_cidrs
+            self.internal_network_updated(
+                interface_name, ip_cidrs, p['mtu'])
+            enable_ra = enable_ra or self._port_has_ipv6_subnet(p)
 
         # Check if there is any pd prefix update
         for p in internal_ports:
             if p['id'] in (set(current_port_ids) & set(existing_port_ids)):
                 for subnet in p.get('subnets', []):
                     if ipv6_utils.is_ipv6_pd_enabled(subnet):
-                        old_prefix = pd.update_subnet(self.router_id,
+                        old_prefix = self.agent.pd.update_subnet(
+                                                      self.router_id,
                                                       subnet['id'],
                                                       subnet['cidr'])
                         if old_prefix:
@@ -549,6 +631,7 @@ class RouterInfo(object):
                                                            subnet['cidr'],
                                                            old_prefix,
                                                            updated_cidrs)
+                            self.pd_subnets[subnet['id']] = subnet['cidr']
                             enable_ra = True
 
         # Enable RA
@@ -564,7 +647,7 @@ class RouterInfo(object):
         for stale_dev in stale_devs:
             LOG.debug('Deleting stale internal router device: %s',
                       stale_dev)
-            pd.remove_stale_ri_ifname(self.router_id, stale_dev)
+            self.agent.pd.remove_stale_ri_ifname(self.router_id, stale_dev)
             self.driver.unplug(stale_dev,
                                namespace=self.ns_name,
                                prefix=INTERNAL_DEV_PREFIX)
@@ -582,7 +665,6 @@ class RouterInfo(object):
                          ex_gw_port['id'],
                          interface_name,
                          ex_gw_port['mac_address'],
-                         bridge=self.agent_conf.external_network_bridge,
                          namespace=ns_name,
                          prefix=EXTERNAL_DEV_PREFIX,
                          mtu=ex_gw_port.get('mtu'))
@@ -614,13 +696,30 @@ class RouterInfo(object):
                 device = ip_lib.IPDevice(device_name, namespace=namespace)
                 device.route.add_route(subnet['gateway_ip'], scope='link')
 
-    def _enable_ra_on_gw(self, ex_gw_port, ns_name, interface_name):
-        gateway_ips = self._get_external_gw_ips(ex_gw_port)
-        if not self.use_ipv6 or self.is_v6_gateway_set(gateway_ips):
+    def _configure_ipv6_params_on_gw(self, ex_gw_port, ns_name, interface_name,
+                                     enabled):
+        if not self.use_ipv6:
             return
 
-        # There is no IPv6 gw_ip, use RouterAdvt for default route.
-        self.driver.configure_ipv6_ra(ns_name, interface_name)
+        disable_ra = not enabled
+        if enabled:
+            gateway_ips = self._get_external_gw_ips(ex_gw_port)
+            if not self.is_v6_gateway_set(gateway_ips):
+                # There is no IPv6 gw_ip, use RouterAdvt for default route.
+                self.driver.configure_ipv6_ra(
+                    ns_name, interface_name, n_const.ACCEPT_RA_WITH_FORWARDING)
+            else:
+                # Otherwise, disable it
+                disable_ra = True
+        if disable_ra:
+            self.driver.configure_ipv6_ra(ns_name, interface_name,
+                                          n_const.ACCEPT_RA_DISABLED)
+        self.driver.configure_ipv6_forwarding(ns_name, interface_name, enabled)
+        # This will make sure the 'all' setting is the same as the interface,
+        # which is needed for forwarding to work.  Don't disable once it's
+        # been enabled so as to not send spurious MLDv2 packets out.
+        if enabled:
+            self.driver.configure_ipv6_forwarding(ns_name, 'all', enabled)
 
     def _external_gateway_added(self, ex_gw_port, interface_name,
                                 ns_name, preserve_ips):
@@ -656,13 +755,13 @@ class RouterInfo(object):
         for ip in gateway_ips:
             device.route.add_gateway(ip)
 
-        self._enable_ra_on_gw(ex_gw_port, ns_name, interface_name)
+        self._configure_ipv6_params_on_gw(ex_gw_port, ns_name, interface_name,
+                                          True)
 
         for fixed_ip in ex_gw_port['fixed_ips']:
             ip_lib.send_ip_addr_adv_notif(ns_name,
                                           interface_name,
-                                          fixed_ip['ip_address'],
-                                          self.agent_conf)
+                                          fixed_ip['ip_address'])
 
     def is_v6_gateway_set(self, gateway_ips):
         """Check to see if list of gateway_ips has an IPv6 gateway.
@@ -673,12 +772,16 @@ class RouterInfo(object):
                    for gw_ip in gateway_ips)
 
     def external_gateway_added(self, ex_gw_port, interface_name):
-        preserve_ips = self._list_floating_ip_cidrs()
+        preserve_ips = self._list_floating_ip_cidrs() + list(
+            self.centralized_port_forwarding_fip_set)
+        preserve_ips.extend(self.agent.pd.get_preserve_ips(self.router_id))
         self._external_gateway_added(
             ex_gw_port, interface_name, self.ns_name, preserve_ips)
 
     def external_gateway_updated(self, ex_gw_port, interface_name):
-        preserve_ips = self._list_floating_ip_cidrs()
+        preserve_ips = self._list_floating_ip_cidrs() + list(
+            self.centralized_port_forwarding_fip_set)
+        preserve_ips.extend(self.agent.pd.get_preserve_ips(self.router_id))
         self._external_gateway_added(
             ex_gw_port, interface_name, self.ns_name, preserve_ips)
 
@@ -693,7 +796,6 @@ class RouterInfo(object):
                                                 ip_addr['ip_address'],
                                                 prefixlen))
         self.driver.unplug(interface_name,
-                           bridge=self.agent_conf.external_network_bridge,
                            namespace=self.ns_name,
                            prefix=EXTERNAL_DEV_PREFIX)
 
@@ -701,7 +803,19 @@ class RouterInfo(object):
     def _gateway_ports_equal(port1, port2):
         return port1 == port2
 
-    def _process_external_gateway(self, ex_gw_port, pd):
+    def _delete_stale_external_devices(self, interface_name):
+        existing_devices = self._get_existing_devices()
+        stale_devs = [dev for dev in existing_devices
+                      if dev.startswith(EXTERNAL_DEV_PREFIX) and
+                      dev != interface_name]
+        for stale_dev in stale_devs:
+            LOG.debug('Deleting stale external router device: %s', stale_dev)
+            self.agent.pd.remove_gw_interface(self.router['id'])
+            self.driver.unplug(stale_dev,
+                               namespace=self.ns_name,
+                               prefix=EXTERNAL_DEV_PREFIX)
+
+    def _process_external_gateway(self, ex_gw_port):
         # TODO(Carl) Refactor to clarify roles of ex_gw_port vs self.ex_gw_port
         ex_gw_port_id = (ex_gw_port and ex_gw_port['id'] or
                          self.ex_gw_port and self.ex_gw_port['id'])
@@ -712,28 +826,19 @@ class RouterInfo(object):
         if ex_gw_port:
             if not self.ex_gw_port:
                 self.external_gateway_added(ex_gw_port, interface_name)
-                pd.add_gw_interface(self.router['id'], interface_name)
+                self.agent.pd.add_gw_interface(self.router['id'],
+                                               interface_name)
             elif not self._gateway_ports_equal(ex_gw_port, self.ex_gw_port):
                 self.external_gateway_updated(ex_gw_port, interface_name)
         elif not ex_gw_port and self.ex_gw_port:
             self.external_gateway_removed(self.ex_gw_port, interface_name)
-            pd.remove_gw_interface(self.router['id'])
+            self.agent.pd.remove_gw_interface(self.router['id'])
         elif not ex_gw_port and not self.ex_gw_port:
             for p in self.internal_ports:
                 interface_name = self.get_internal_device_name(p['id'])
                 self.gateway_redirect_cleanup(interface_name)
 
-        existing_devices = self._get_existing_devices()
-        stale_devs = [dev for dev in existing_devices
-                      if dev.startswith(EXTERNAL_DEV_PREFIX)
-                      and dev != interface_name]
-        for stale_dev in stale_devs:
-            LOG.debug('Deleting stale external router device: %s', stale_dev)
-            pd.remove_gw_interface(self.router['id'])
-            self.driver.unplug(stale_dev,
-                               bridge=self.agent_conf.external_network_bridge,
-                               namespace=self.ns_name,
-                               prefix=EXTERNAL_DEV_PREFIX)
+        self._delete_stale_external_devices(interface_name)
 
         # Process SNAT rules for external gateway
         gw_port = self._router.get('gw_port')
@@ -751,19 +856,21 @@ class RouterInfo(object):
             self._prevent_snat_for_internal_traffic_rule(interface_name))
         # Makes replies come back through the router to reverse DNAT
         ext_in_mark = self.agent_conf.external_ingress_mark
-        snat_internal_traffic_to_floating_ip = (
-            'snat', '-m mark ! --mark %s/%s '
-                    '-m conntrack --ctstate DNAT '
-                    '-j SNAT --to-source %s'
-                    % (ext_in_mark, n_const.ROUTER_MARK_MASK, ex_gw_ip))
+        to_source = ('-m mark ! --mark %s/%s '
+                     '-m conntrack --ctstate DNAT '
+                     '-j SNAT --to-source %s'
+                     % (ext_in_mark, lib_constants.ROUTER_MARK_MASK, ex_gw_ip))
+        if self.iptables_manager.random_fully:
+            to_source += ' --random-fully'
+        snat_internal_traffic_to_floating_ip = ('snat', to_source)
         return [dont_snat_traffic_to_internal_ports_if_not_to_floating_ip,
                 snat_internal_traffic_to_floating_ip]
 
     def external_gateway_nat_snat_rules(self, ex_gw_ip, interface_name):
-        snat_normal_external_traffic = (
-            'snat', '-o %s -j SNAT --to-source %s' %
-                    (interface_name, ex_gw_ip))
-        return [snat_normal_external_traffic]
+        to_source = '-o %s -j SNAT --to-source %s' % (interface_name, ex_gw_ip)
+        if self.iptables_manager.random_fully:
+            to_source += ' --random-fully'
+        return [('snat', to_source)]
 
     def external_gateway_mangle_rules(self, interface_name):
         mark = self.agent_conf.external_ingress_mark
@@ -813,11 +920,11 @@ class RouterInfo(object):
                              self.iptables_manager,
                              interface_name)
 
-    def _process_external_on_delete(self, agent):
+    def _process_external_on_delete(self):
         fip_statuses = {}
         try:
             ex_gw_port = self.get_ex_gw_port()
-            self._process_external_gateway(ex_gw_port, agent.pd)
+            self._process_external_gateway(ex_gw_port)
             if not ex_gw_port:
                 return
 
@@ -825,19 +932,19 @@ class RouterInfo(object):
                 ex_gw_port)
             fip_statuses = self.configure_fip_addresses(interface_name)
 
-        except (n_exc.FloatingIpSetupException):
-                # All floating IPs must be put in error state
-                LOG.exception(_LE("Failed to process floating IPs."))
-                fip_statuses = self.put_fips_in_error_state()
+        except l3_exc.FloatingIpSetupException:
+            # All floating IPs must be put in error state
+            LOG.exception("Failed to process floating IPs.")
+            fip_statuses = self.put_fips_in_error_state()
         finally:
-            self.update_fip_statuses(agent, fip_statuses)
+            self.update_fip_statuses(fip_statuses)
 
-    def process_external(self, agent):
+    def process_external(self):
         fip_statuses = {}
         try:
             with self.iptables_manager.defer_apply():
                 ex_gw_port = self.get_ex_gw_port()
-                self._process_external_gateway(ex_gw_port, agent.pd)
+                self._process_external_gateway(ex_gw_port)
                 if not ex_gw_port:
                     return
 
@@ -850,15 +957,15 @@ class RouterInfo(object):
                 ex_gw_port)
             fip_statuses = self.configure_fip_addresses(interface_name)
 
-        except (n_exc.FloatingIpSetupException,
-                n_exc.IpTablesApplyException):
-                # All floating IPs must be put in error state
-                LOG.exception(_LE("Failed to process floating IPs."))
-                fip_statuses = self.put_fips_in_error_state()
+        except (l3_exc.FloatingIpSetupException,
+                l3_exc.IpTablesApplyException):
+            # All floating IPs must be put in error state
+            LOG.exception("Failed to process floating IPs.")
+            fip_statuses = self.put_fips_in_error_state()
         finally:
-            self.update_fip_statuses(agent, fip_statuses)
+            self.update_fip_statuses(fip_statuses)
 
-    def update_fip_statuses(self, agent, fip_statuses):
+    def update_fip_statuses(self, fip_statuses):
         # Identify floating IPs which were disabled
         existing_floating_ips = self.floating_ips
         self.floating_ips = set(fip_statuses.keys())
@@ -871,8 +978,8 @@ class RouterInfo(object):
             return
         LOG.debug('Sending floating ip statuses: %s', fip_statuses)
         # Update floating IP status on the neutron server
-        agent.plugin_rpc.update_floatingip_statuses(
-            agent.context, self.router_id, fip_statuses)
+        self.agent.plugin_rpc.update_floatingip_statuses(
+            self.agent.context, self.router_id, fip_statuses)
 
     def initialize_address_scope_iptables(self):
         self._initialize_address_scope_iptables(self.iptables_manager)
@@ -921,6 +1028,21 @@ class RouterInfo(object):
         iptables_manager.ipv6['mangle'].add_rule(
             'PREROUTING', copy_address_scope_for_existing)
 
+    def initialize_metadata_iptables(self):
+        # Always mark incoming metadata requests, that way any stray
+        # requests that arrive before the filter metadata redirect
+        # rule is installed will be dropped.
+        mark_metadata_for_internal_interfaces = (
+            '-d 169.254.169.254/32 '
+            '-i %(interface_name)s '
+            '-p tcp -m tcp --dport 80 '
+            '-j MARK --set-xmark %(value)s/%(mask)s' %
+            {'interface_name': INTERNAL_DEV_PREFIX + '+',
+             'value': self.agent_conf.metadata_access_mark,
+             'mask': n_const.ROUTER_MARK_MASK})
+        self.iptables_manager.ipv4['mangle'].add_rule(
+            'PREROUTING', mark_metadata_for_internal_interfaces)
+
     def _get_port_devicename_scopemark(self, ports, name_generator):
         devicename_scopemark = {lib_constants.IP_VERSION_4: dict(),
                                 lib_constants.IP_VERSION_6: dict()}
@@ -928,7 +1050,7 @@ class RouterInfo(object):
             device_name = name_generator(p['id'])
             ip_cidrs = common_utils.fixed_ip_cidrs(p['fixed_ips'])
             port_as_marks = self.get_port_address_scope_mark(p)
-            for ip_version in {ip_lib.get_ip_version(cidr)
+            for ip_version in {common_utils.get_ip_version(cidr)
                                for cidr in ip_cidrs}:
                 devicename_scopemark[ip_version][device_name] = (
                     port_as_marks[ip_version])
@@ -966,8 +1088,8 @@ class RouterInfo(object):
             iptables = iptables_manager.get_tables(ip_version)
             iptables['mangle'].empty_chain('scope')
             iptables['filter'].empty_chain('scope')
-            dont_block_external = (ip_version == lib_constants.IP_VERSION_4
-                                   and self._snat_enabled and external_port)
+            dont_block_external = (ip_version == lib_constants.IP_VERSION_4 and
+                                   self._snat_enabled and external_port)
             for device_name, mark in scopemarks.items():
                 # Add address scope iptables rule
                 iptables['mangle'].add_rule(
@@ -978,6 +1100,9 @@ class RouterInfo(object):
                 iptables['filter'].add_rule(
                     'scope',
                     self.address_scope_filter_rule(device_name, mark))
+        for subnet_id, prefix in self.pd_subnets.items():
+            if prefix != lib_constants.PROVISIONAL_IPV6_PD_PREFIX:
+                self._process_pd_iptables_rules(prefix, subnet_id)
 
     def process_ports_address_scope_iptables(self):
         ports_scopemark = self._get_address_scope_mark()
@@ -1028,7 +1153,7 @@ class RouterInfo(object):
             self.process_floating_ip_address_scope_rules()
 
     @common_utils.exception_logger()
-    def process_delete(self, agent):
+    def process_delete(self):
         """Process the delete of this router
 
         This method is the point where the agent requests that this router
@@ -1038,17 +1163,17 @@ class RouterInfo(object):
 
         :param agent: Passes the agent in order to send RPC messages.
         """
-        LOG.debug("process router delete")
+        LOG.debug("Process delete, router %s", self.router['id'])
         if self.router_namespace.exists():
-            self._process_internal_ports(agent.pd)
-            agent.pd.sync_router(self.router['id'])
-            self._process_external_on_delete(agent)
+            self._process_internal_ports()
+            self.agent.pd.sync_router(self.router['id'])
+            self._process_external_on_delete()
         else:
-            LOG.warning(_LW("Can't gracefully delete the router %s: "
-                            "no router namespace found."), self.router['id'])
+            LOG.warning("Can't gracefully delete the router %s: "
+                        "no router namespace found", self.router['id'])
 
     @common_utils.exception_logger()
-    def process(self, agent):
+    def process(self):
         """Process updates to this router
 
         This method is the point where the agent requests that updates be
@@ -1056,19 +1181,21 @@ class RouterInfo(object):
 
         :param agent: Passes the agent in order to send RPC messages.
         """
-        LOG.debug("process router updates")
-        self._process_internal_ports(agent.pd)
-        agent.pd.sync_router(self.router['id'])
-        self.process_external(agent)
+        LOG.debug("Process updates, router %s", self.router['id'])
+        self.centralized_port_forwarding_fip_set = set(self.router.get(
+            'port_forwardings_fip_set', set()))
+        self._process_internal_ports()
+        self.agent.pd.sync_router(self.router['id'])
+        self.process_external()
         self.process_address_scope()
         # Process static routes for router
         self.routes_updated(self.routes, self.router['routes'])
         self.routes = self.router['routes']
 
-        # Update ex_gw_port and enable_snat on the router info cache
+        # Update ex_gw_port on the router info cache
         self.ex_gw_port = self.get_ex_gw_port()
-        self.fip_map = dict([(fip['floating_ip_address'],
-                              fip['fixed_ip_address'])
-                             for fip in self.get_floating_ips()])
-        # TODO(Carl) FWaaS uses this.  Why is it set after processing is done?
-        self.enable_snat = self.router.get('enable_snat')
+        self.fip_map = dict((fip['floating_ip_address'],
+                             fip['fixed_ip_address'])
+                            for fip in self.get_floating_ips())
+        self.fip_managed_by_port_forwardings = self.router.get(
+            'fip_managed_by_port_forwardings')
