@@ -20,7 +20,12 @@ import socket
 import sys
 import time
 
+from neutron_lib.agent import topics
+from neutron_lib.api.definitions import portbindings
 from neutron_lib import constants as n_constants
+from neutron_lib import context
+from neutron_lib.placement import utils as place_utils
+from neutron_lib.utils import helpers
 from oslo_config import cfg
 from oslo_log import log as logging
 import oslo_messaging
@@ -28,17 +33,16 @@ from oslo_service import loopingcall
 from osprofiler import profiler
 import six
 
-from neutron._i18n import _, _LE, _LI, _LW
+from neutron._i18n import _
 from neutron.agent.l2 import l2_agent_extensions_manager as ext_manager
 from neutron.agent import rpc as agent_rpc
-from neutron.agent import securitygroups_rpc as sg_rpc
+from neutron.agent import securitygroups_rpc as agent_sg_rpc
 from neutron.api.rpc.callbacks import resources
+from neutron.api.rpc.handlers import securitygroups_rpc as sg_rpc
 from neutron.common import config as common_config
+from neutron.common import constants as c_const
 from neutron.common import profiler as setup_profiler
-from neutron.common import topics
 from neutron.common import utils as n_utils
-from neutron import context
-from neutron.extensions import portbindings
 from neutron.plugins.ml2.drivers.mech_sriov.agent.common import config
 from neutron.plugins.ml2.drivers.mech_sriov.agent.common \
     import exceptions as exc
@@ -57,8 +61,9 @@ class SriovNicSwitchRpcCallbacks(sg_rpc.SecurityGroupAgentRpcCallbackMixin):
     #   1.3 Added param devices_to_update to security_groups_provider_updated
     #       (works with NoopFirewallDriver)
     #   1.4 Added support for network_update
+    #   1.5 Added support for binding_activate and binding_deactivate
 
-    target = oslo_messaging.Target(version='1.4')
+    target = oslo_messaging.Target(version='1.5')
 
     def __init__(self, context, agent, sg_agent):
         super(SriovNicSwitchRpcCallbacks, self).__init__()
@@ -104,34 +109,62 @@ class SriovNicSwitchRpcCallbacks(sg_rpc.SecurityGroupAgentRpcCallbackMixin):
         for port_data in self.agent.network_ports[network_id]:
             self.agent.updated_devices.add(port_data['device'])
 
+    def binding_activate(self, context, **kwargs):
+        if kwargs.get('host') != self.agent.conf.host:
+            return
+        LOG.debug("binding activate for port %s", kwargs.get('port_id'))
+        device_details = self.agent.get_device_details_from_port_id(
+            kwargs.get('port_id'))
+        mac = device_details.get('mac_address')
+        binding_profile = device_details.get('profile')
+        if binding_profile:
+            pci_slot = binding_profile.get('pci_slot')
+            self.agent.activated_bindings.add((mac, pci_slot))
+        else:
+            LOG.warning("binding_profile not found for port %s.",
+                        kwargs.get('port_id'))
+
+    def binding_deactivate(self, context, **kwargs):
+        if kwargs.get('host') != self.agent.conf.host:
+            return
+        LOG.debug("binding deactivate for port %s. NOOP.",
+                  kwargs.get('port_id'))
+
 
 @profiler.trace_cls("rpc")
 class SriovNicSwitchAgent(object):
     def __init__(self, physical_devices_mappings, exclude_devices,
-                 polling_interval):
+                 polling_interval, rp_bandwidths, rp_inventory_defaults):
 
         self.polling_interval = polling_interval
         self.network_ports = collections.defaultdict(list)
         self.conf = cfg.CONF
+        self.device_mappings = physical_devices_mappings
+        self.exclude_devices = exclude_devices
         self.setup_eswitch_mgr(physical_devices_mappings,
                                exclude_devices)
 
         # Stores port update notifications for processing in the main loop
         self.updated_devices = set()
+        # Stores <mac, pci_slot> pairs for ports whose binding has been
+        # activated.
+        self.activated_bindings = set()
 
         self.context = context.get_admin_context_without_session()
         self.plugin_rpc = agent_rpc.PluginApi(topics.PLUGIN)
         self.sg_plugin_rpc = sg_rpc.SecurityGroupServerRpcApi(topics.PLUGIN)
-        self.sg_agent = sg_rpc.SecurityGroupAgentRpc(self.context,
-                self.sg_plugin_rpc)
+        self.sg_agent = agent_sg_rpc.SecurityGroupAgentRpc(
+            self.context, self.sg_plugin_rpc)
         self._setup_rpc()
         self.ext_manager = self._create_agent_extension_manager(
             self.connection)
 
         configurations = {'device_mappings': physical_devices_mappings,
+                          c_const.RP_BANDWIDTHS: rp_bandwidths,
+                          c_const.RP_INVENTORY_DEFAULTS: rp_inventory_defaults,
                           'extensions': self.ext_manager.names()}
 
-        #TODO(mangelajo): optimize resource_versions (see ovs agent)
+        # TODO(mangelajo): optimize resource_versions (see ovs agent)
         self.agent_state = {
             'binary': 'neutron-sriov-nic-agent',
             'host': self.conf.host,
@@ -148,9 +181,9 @@ class SriovNicSwitchAgent(object):
 
     def _setup_rpc(self):
         self.agent_id = 'nic-switch-agent.%s' % socket.gethostname()
-        LOG.info(_LI("RPC agent_id: %s"), self.agent_id)
-
+        LOG.info("RPC agent_id: %s", self.agent_id)
         self.topic = topics.AGENT
+        self.failed_report_state = False
         self.state_rpc = agent_rpc.PluginReportStateAPI(topics.REPORTS)
         # RPC network init
         # Handle updates from service
@@ -159,7 +192,9 @@ class SriovNicSwitchAgent(object):
         # Define the listening consumers for the agent
         consumers = [[topics.PORT, topics.UPDATE],
                      [topics.NETWORK, topics.UPDATE],
-                     [topics.SECURITY_GROUP, topics.UPDATE]]
+                     [topics.SECURITY_GROUP, topics.UPDATE],
+                     [topics.PORT_BINDING, topics.DEACTIVATE],
+                     [topics.PORT_BINDING, topics.ACTIVATE]]
         self.connection = agent_rpc.create_consumers(self.endpoints,
                                                      self.topic,
                                                      consumers,
@@ -173,8 +208,6 @@ class SriovNicSwitchAgent(object):
 
     def _report_state(self):
         try:
-            devices = len(self.eswitch_mgr.get_assigned_devices_info())
-            self.agent_state.get('configurations')['devices'] = devices
             self.state_rpc.report_state(self.context,
                                         self.agent_state)
 
@@ -182,7 +215,12 @@ class SriovNicSwitchAgent(object):
             self.agent_state.pop('resource_versions', None)
             self.agent_state.pop('start_flag', None)
         except Exception:
-            LOG.exception(_LE("Failed reporting state!"))
+            self.failed_report_state = True
+            LOG.exception("Failed reporting state!")
+            return
+        if self.failed_report_state:
+            self.failed_report_state = False
+            LOG.info("Successfully reported state after a previous failure.")
 
     def _create_agent_extension_manager(self, connection):
         ext_manager.register_opts(self.conf)
@@ -197,6 +235,7 @@ class SriovNicSwitchAgent(object):
 
     def scan_devices(self, registered_devices, updated_devices):
         curr_devices = self.eswitch_mgr.get_assigned_devices_info()
+        self.agent_state.get('configurations')['devices'] = len(curr_devices)
         device_info = {}
         device_info['current'] = curr_devices
         device_info['added'] = curr_devices - registered_devices
@@ -208,9 +247,9 @@ class SriovNicSwitchAgent(object):
         return device_info
 
     def _device_info_has_changes(self, device_info):
-        return (device_info.get('added')
-                or device_info.get('updated')
-                or device_info.get('removed'))
+        return (device_info.get('added') or
+                device_info.get('updated') or
+                device_info.get('removed'))
 
     def process_network_devices(self, device_info):
         resync_a = False
@@ -223,8 +262,8 @@ class SriovNicSwitchAgent(object):
         # Updated devices are processed the same as new ones, as their
         # admin_state_up may have changed. The set union prevents duplicating
         # work when a device is new and updated in the same polling iteration.
-        devices_added_updated = (set(device_info.get('added'))
-                                 | set(device_info.get('updated')))
+        devices_added_updated = (set(device_info.get('added')) |
+                                 set(device_info.get('updated')))
         if devices_added_updated:
             resync_a = self.treat_devices_added_updated(devices_added_updated)
 
@@ -233,39 +272,32 @@ class SriovNicSwitchAgent(object):
         # If one of the above operations fails => resync with plugin
         return (resync_a | resync_b)
 
-    def treat_device(self, device, pci_slot, admin_state_up, spoofcheck=True):
+    def treat_device(self, device, pci_slot, admin_state_up, spoofcheck=True,
+                     propagate_uplink_state=False):
         if self.eswitch_mgr.device_exists(device, pci_slot):
             try:
                 self.eswitch_mgr.set_device_spoofcheck(device, pci_slot,
                                                        spoofcheck)
             except Exception:
-                LOG.warning(_LW("Failed to set spoofcheck for device %s"),
+                LOG.warning("Failed to set spoofcheck for device %s",
                             device)
-            LOG.info(_LI("Device %(device)s spoofcheck %(spoofcheck)s"),
+            LOG.info("Device %(device)s spoofcheck %(spoofcheck)s",
                      {"device": device, "spoofcheck": spoofcheck})
 
             try:
                 self.eswitch_mgr.set_device_state(device, pci_slot,
-                                                  admin_state_up)
+                                                  admin_state_up,
+                                                  propagate_uplink_state)
             except exc.IpCommandOperationNotSupportedError:
-                LOG.warning(_LW("Device %s does not support state change"),
+                LOG.warning("Device %s does not support state change",
                             device)
             except exc.SriovNicError:
-                LOG.warning(_LW("Failed to set device %s state"), device)
-                return
-            if admin_state_up:
-                # update plugin about port status
-                self.plugin_rpc.update_device_up(self.context,
-                                                 device,
-                                                 self.agent_id,
-                                                 cfg.CONF.host)
-            else:
-                self.plugin_rpc.update_device_down(self.context,
-                                                   device,
-                                                   self.agent_id,
-                                                   cfg.CONF.host)
+                LOG.warning("Failed to set device %s state", device)
+                return False
         else:
-            LOG.info(_LI("No device with MAC %s defined on agent."), device)
+            LOG.info("No device with MAC %s defined on agent.", device)
+            return False
+        return True
 
     def _update_network_ports(self, network_id, port_id, mac_pci_slot):
         self._clean_network_ports(mac_pci_slot)
@@ -274,7 +306,7 @@ class SriovNicSwitchAgent(object):
             "device": mac_pci_slot})
 
     def _clean_network_ports(self, mac_pci_slot):
-        for netid, ports_list in six.iteritems(self.network_ports):
+        for netid, ports_list in self.network_ports.items():
             for port_data in ports_list:
                 if mac_pci_slot == port_data['device']:
                     ports_list.remove(port_data)
@@ -286,7 +318,7 @@ class SriovNicSwitchAgent(object):
         try:
             macs_list = set([device_info[0] for device_info in devices_info])
             devices_details_list = self.plugin_rpc.get_devices_details_list(
-                self.context, macs_list, self.agent_id)
+                self.context, macs_list, self.agent_id, self.conf.host)
         except Exception as e:
             LOG.debug("Unable to get port details for devices "
                       "with MAC addresses %(devices)s: %(e)s",
@@ -294,35 +326,57 @@ class SriovNicSwitchAgent(object):
             # resync is needed
             return True
 
+        devices_up = set()
+        devices_down = set()
+        resync = False
         for device_details in devices_details_list:
             device = device_details['device']
             LOG.debug("Port with MAC address %s is added", device)
 
             if 'port_id' in device_details:
-                LOG.info(_LI("Port %(device)s updated. Details: %(details)s"),
+                LOG.info("Port %(device)s updated. Details: %(details)s",
                          {'device': device, 'details': device_details})
                 port_id = device_details['port_id']
                 profile = device_details['profile']
                 spoofcheck = device_details.get('port_security_enabled', True)
-                self.treat_device(device,
-                                  profile.get('pci_slot'),
-                                  device_details['admin_state_up'],
-                                  spoofcheck)
+                if self.treat_device(
+                        device,
+                        profile.get('pci_slot'),
+                        device_details['admin_state_up'],
+                        spoofcheck,
+                        device_details['propagate_uplink_status']):
+                    if device_details['admin_state_up']:
+                        devices_up.add(device)
+                    else:
+                        devices_down.add(device)
+                else:
+                    resync = True
                 self._update_network_ports(device_details['network_id'],
                                            port_id,
                                            (device, profile.get('pci_slot')))
                 self.ext_manager.handle_port(self.context, device_details)
-            else:
-                LOG.info(_LI("Device with MAC %s not defined on plugin"),
+            elif c_const.NO_ACTIVE_BINDING in device_details:
+                # Port was added but its binding in this agent
+                # hasn't been activated yet. It will be treated as
+                # added when binding is activated
+                LOG.info("Device with MAC %s has no active binding in host",
                          device)
-        return False
+            else:
+                LOG.info("Device with MAC %s not defined on plugin",
+                         device)
+        self.plugin_rpc.update_device_list(self.context,
+                                           devices_up,
+                                           devices_down,
+                                           self.agent_id,
+                                           self.conf.host)
+        return resync
 
     def treat_devices_removed(self, devices):
         resync = False
         for device in devices:
             mac, pci_slot = device
-            LOG.info(_LI("Removing device with MAC address %(mac)s and "
-                         "PCI slot %(pci_slot)s"),
+            LOG.info("Removing device with MAC address %(mac)s and "
+                     "PCI slot %(pci_slot)s",
                      {'mac': mac, 'pci_slot': pci_slot})
             try:
                 port_id = self._clean_network_ports(device)
@@ -332,8 +386,8 @@ class SriovNicSwitchAgent(object):
                             'profile': {'pci_slot': pci_slot}}
                     self.ext_manager.delete_port(self.context, port)
                 else:
-                    LOG.warning(_LW("port_id to device with MAC "
-                                    "%s not found"), mac)
+                    LOG.warning("port_id to device with MAC "
+                                "%s not found", mac)
                 dev_details = self.plugin_rpc.update_device_down(self.context,
                                                                  mac,
                                                                  self.agent_id,
@@ -346,8 +400,8 @@ class SriovNicSwitchAgent(object):
                 resync = True
                 continue
             if dev_details['exists']:
-                LOG.info(_LI("Port with MAC %(mac)s and PCI slot "
-                             "%(pci_slot)s updated."),
+                LOG.info("Port with MAC %(mac)s and PCI slot "
+                         "%(pci_slot)s updated.",
                          {'mac': mac, 'pci_slot': pci_slot})
             else:
                 LOG.debug("Device with MAC %(mac)s and PCI slot "
@@ -355,18 +409,48 @@ class SriovNicSwitchAgent(object):
                           {'mac': mac, 'pci_slot': pci_slot})
         return resync
 
+    def process_activated_bindings(self, device_info, activated_bindings_copy):
+        """Process activated bindings.
+
+        Add activated bindings to the 'added' set in device info.
+
+        :param device_info: A dict that contains the set of 'current', 'added',
+                            'removed' and 'updated' ports.
+        :param activated_bindings_copy: A set of activated port bindings.
+        :return: None
+        """
+        LOG.debug("Processing activated bindings: %s", activated_bindings_copy)
+        # Compute which ports for activated bindings are already present
+        activated_bindings_copy &= device_info['current']
+        # Treat them as just added
+        device_info['added'] |= activated_bindings_copy
+
+    def get_device_details_from_port_id(self, port_id):
+        """Get device details from server
+
+        :param port_id: Port identifier (UUID).
+        :return: A dict containing various port attributes if the port is
+                 bound to the host. In case the port is not bound to the host
+                 then the method will return A dict with a minimal set of
+                 attributes e.g  {'device': port_id}.
+        """
+        return self.plugin_rpc.get_device_details(self.context,
+                                                  port_id,
+                                                  self.agent_id,
+                                                  host=cfg.CONF.host)
+
     def daemon_loop(self):
         sync = True
         devices = set()
 
-        LOG.info(_LI("SRIOV NIC Agent RPC Daemon Started!"))
+        LOG.info("SRIOV NIC Agent RPC Daemon Started!")
 
         while True:
             start = time.time()
             LOG.debug("Agent rpc_loop - iteration:%d started",
                       self.iter_num)
             if sync:
-                LOG.info(_LI("Agent out of sync with plugin!"))
+                LOG.info("Agent out of sync with plugin!")
                 devices.clear()
                 sync = False
             device_info = {}
@@ -376,8 +460,15 @@ class SriovNicSwitchAgent(object):
             # two statements, this will should be thread-safe.
             updated_devices_copy = self.updated_devices
             self.updated_devices = set()
+            activated_bindings_copy = self.activated_bindings
+            self.activated_bindings = set()
             try:
+                self.eswitch_mgr.discover_devices(self.device_mappings,
+                                                  self.exclude_devices)
                 device_info = self.scan_devices(devices, updated_devices_copy)
+                if activated_bindings_copy:
+                    self.process_activated_bindings(device_info,
+                                                    activated_bindings_copy)
                 if self._device_info_has_changes(device_info):
                     LOG.debug("Agent loop found changes! %s", device_info)
                     # If treat devices fails - indicates must resync with
@@ -385,12 +476,13 @@ class SriovNicSwitchAgent(object):
                     sync = self.process_network_devices(device_info)
                     devices = device_info['current']
             except Exception:
-                LOG.exception(_LE("Error in agent loop. Devices info: %s"),
+                LOG.exception("Error in agent loop. Devices info: %s",
                               device_info)
                 sync = True
                 # Restore devices that were removed from this set earlier
                 # without overwriting ones that may have arrived since.
                 self.updated_devices |= updated_devices_copy
+                self.activated_bindings |= activated_bindings_copy
 
             # sleep till end of polling interval
             elapsed = (time.time() - start)
@@ -414,25 +506,33 @@ class SriovNicAgentConfigParser(object):
 
         Parse and validate the consistency in both mappings
         """
-        self.device_mappings = n_utils.parse_mappings(
+        self.device_mappings = helpers.parse_mappings(
             cfg.CONF.SRIOV_NIC.physical_device_mappings, unique_keys=False)
         self.exclude_devices = config.parse_exclude_devices(
             cfg.CONF.SRIOV_NIC.exclude_devices)
+        self.rp_bandwidths = place_utils.parse_rp_bandwidths(
+            cfg.CONF.SRIOV_NIC.resource_provider_bandwidths)
+        self.rp_inventory_defaults = place_utils.parse_rp_inventory_defaults(
+            cfg.CONF.SRIOV_NIC.resource_provider_inventory_defaults)
         self._validate()
 
     def _validate(self):
         """Validate configuration.
 
         Validate that network_device in excluded_device
-        exists in device mappings
+        exists in device mappings.
+        Validate that network_device in resource_provider_bandwidths
+        exists in device mappings.
         """
         dev_net_set = set(itertools.chain.from_iterable(
                           six.itervalues(self.device_mappings)))
         for dev_name in self.exclude_devices.keys():
             if dev_name not in dev_net_set:
-                raise ValueError(_("Device name %(dev_name)s is missing from "
-                                   "physical_device_mappings") % {'dev_name':
-                                                                  dev_name})
+                raise ValueError(_(
+                    "Invalid exclude_devices: "
+                    "Device name %(dev_name)s is missing from "
+                    "physical_device_mappings") % {'dev_name': dev_name})
+        n_utils.validate_rp_bandwidth(self.rp_bandwidths, dev_net_set)
 
 
 def main():
@@ -444,23 +544,27 @@ def main():
         config_parser.parse()
         device_mappings = config_parser.device_mappings
         exclude_devices = config_parser.exclude_devices
+        rp_bandwidths = config_parser.rp_bandwidths
+        rp_inventory_defaults = config_parser.rp_inventory_defaults
 
     except ValueError:
-        LOG.exception(_LE("Failed on Agent configuration parse. "
-                          "Agent terminated!"))
+        LOG.exception("Failed on Agent configuration parse. "
+                      "Agent terminated!")
         raise SystemExit(1)
-    LOG.info(_LI("Physical Devices mappings: %s"), device_mappings)
-    LOG.info(_LI("Exclude Devices: %s"), exclude_devices)
+    LOG.info("Physical Devices mappings: %s", device_mappings)
+    LOG.info("Exclude Devices: %s", exclude_devices)
 
     polling_interval = cfg.CONF.AGENT.polling_interval
     try:
         agent = SriovNicSwitchAgent(device_mappings,
                                     exclude_devices,
-                                    polling_interval)
+                                    polling_interval,
+                                    rp_bandwidths,
+                                    rp_inventory_defaults)
     except exc.SriovNicError:
-        LOG.exception(_LE("Agent Initialization Failed"))
+        LOG.exception("Agent Initialization Failed")
         raise SystemExit(1)
     # Start everything.
     setup_profiler.setup("neutron-sriov-nic-agent", cfg.CONF.host)
-    LOG.info(_LI("Agent initialized successfully, now running... "))
+    LOG.info("Agent initialized successfully, now running... ")
     agent.daemon_loop()
