@@ -17,11 +17,10 @@ import copy
 
 import mock
 from neutron_lib import constants
-import six
 import testtools
 
+from neutron.agent.common import ovs_lib
 from neutron.agent.l3 import agent as neutron_l3_agent
-from neutron.agent.l3 import namespaces
 from neutron.agent.linux import ip_lib
 from neutron.common import ipv6_utils
 from neutron.common import utils as common_utils
@@ -59,7 +58,7 @@ class L3HATestCase(framework.L3AgentTestFramework):
         # Get the last state reported for each router
         actual_router_states = {}
         for call in calls:
-            for router_id, state in six.iteritems(call):
+            for router_id, state in call.items():
                 actual_router_states[router_id] = state
 
         return actual_router_states == expected
@@ -86,12 +85,14 @@ class L3HATestCase(framework.L3AgentTestFramework):
         self._test_conntrack_disassociate_fip(ha=True)
 
     def test_ipv6_ha_router_lifecycle(self):
-        self._router_lifecycle(enable_ha=True, ip_version=6)
+        self._router_lifecycle(enable_ha=True,
+                               ip_version=constants.IP_VERSION_6)
 
     def test_ipv6_ha_router_lifecycle_with_no_gw_subnet(self):
         self.agent.conf.set_override('ipv6_gateway',
                                      'fe80::f816:3eff:fe2e:1')
-        self._router_lifecycle(enable_ha=True, ip_version=6,
+        self._router_lifecycle(enable_ha=True,
+                               ip_version=constants.IP_VERSION_6,
                                v6_ext_gw_with_sub=False)
 
     def test_ipv6_ha_router_lifecycle_with_no_gw_subnet_for_router_advts(self):
@@ -100,19 +101,58 @@ class L3HATestCase(framework.L3AgentTestFramework):
         self._router_lifecycle(enable_ha=True, dual_stack=True,
                                v6_ext_gw_with_sub=False)
 
-    @testtools.skipUnless(ipv6_utils.is_enabled(), "IPv6 is not enabled")
-    def test_ipv6_router_advts_after_router_state_change(self):
+    def _test_ipv6_router_advts_and_fwd_helper(self, state, enable_v6_gw,
+                                               expected_ra,
+                                               expected_forwarding):
         # Schedule router to l3 agent, and then add router gateway. Verify
-        # that router gw interface is configured to receive Router Advts.
+        # that router gw interface is configured to receive Router Advts and
+        # IPv6 forwarding is enabled.
         router_info = l3_test_common.prepare_router_data(
             enable_snat=True, enable_ha=True, dual_stack=True, enable_gw=False)
         router = self.manage_router(self.agent, router_info)
         common_utils.wait_until_true(lambda: router.ha_state == 'master')
+        if state == 'backup':
+            self.fail_ha_router(router)
+            common_utils.wait_until_true(lambda: router.ha_state == 'backup')
         _ext_dev_name, ex_port = l3_test_common.prepare_ext_gw_test(
-            mock.Mock(), router)
+            mock.Mock(), router, dual_stack=enable_v6_gw)
         router_info['gw_port'] = ex_port
-        router.process(self.agent)
-        self._assert_ipv6_accept_ra(router)
+        router.process()
+        self._assert_ipv6_accept_ra(router, expected_ra)
+        # As router is going first to master and than to backup mode,
+        # ipv6_forwarding should be enabled on "all" interface always after
+        # that transition
+        self._assert_ipv6_forwarding(router, expected_forwarding,
+                                     True)
+
+    @testtools.skipUnless(ipv6_utils.is_enabled_and_bind_by_default(),
+                          "IPv6 is not enabled")
+    def test_ipv6_router_advts_and_fwd_after_router_state_change_master(self):
+        # Check that RA and forwarding are enabled when there's no IPv6
+        # gateway.
+        self._test_ipv6_router_advts_and_fwd_helper('master',
+                                                    enable_v6_gw=False,
+                                                    expected_ra=True,
+                                                    expected_forwarding=True)
+        # Check that RA is disabled and forwarding is enabled when an IPv6
+        # gateway is configured.
+        self._test_ipv6_router_advts_and_fwd_helper('master',
+                                                    enable_v6_gw=True,
+                                                    expected_ra=False,
+                                                    expected_forwarding=True)
+
+    @testtools.skipUnless(ipv6_utils.is_enabled_and_bind_by_default(),
+                          "IPv6 is not enabled")
+    def test_ipv6_router_advts_and_fwd_after_router_state_change_backup(self):
+        # Check that both RA and forwarding are disabled on backup instances
+        self._test_ipv6_router_advts_and_fwd_helper('backup',
+                                                    enable_v6_gw=False,
+                                                    expected_ra=False,
+                                                    expected_forwarding=False)
+        self._test_ipv6_router_advts_and_fwd_helper('backup',
+                                                    enable_v6_gw=True,
+                                                    expected_ra=False,
+                                                    expected_forwarding=False)
 
     def test_keepalived_configuration(self):
         router_info = self.generate_router_info(enable_ha=True)
@@ -137,7 +177,7 @@ class L3HATestCase(framework.L3AgentTestFramework):
         router.router['gw_port']['subnets'] = subnets
         router.router['gw_port']['fixed_ips'] = fixed_ips
 
-        router.process(self.agent)
+        router.process()
 
         # Get the updated configuration and assert that both FIPs are in,
         # and that the GW IP address was updated.
@@ -176,7 +216,8 @@ class L3HATestCase(framework.L3AgentTestFramework):
                 router1.get_ha_device_name()))
 
     def test_ha_router_ipv6_radvd_status(self):
-        router_info = self.generate_router_info(ip_version=6, enable_ha=True)
+        router_info = self.generate_router_info(
+            ip_version=constants.IP_VERSION_6, enable_ha=True)
         router1 = self.manage_router(self.agent, router_info)
         common_utils.wait_until_true(lambda: router1.ha_state == 'master')
         common_utils.wait_until_true(lambda: router1.radvd.enabled)
@@ -202,7 +243,8 @@ class L3HATestCase(framework.L3AgentTestFramework):
         _check_lla_status(router1, False)
 
     def test_ha_router_process_ipv6_subnets_to_existing_port(self):
-        router_info = self.generate_router_info(enable_ha=True, ip_version=6)
+        router_info = self.generate_router_info(enable_ha=True,
+            ip_version=constants.IP_VERSION_6)
         router = self.manage_router(self.agent, router_info)
 
         def verify_ip_in_keepalived_config(router, iface):
@@ -217,9 +259,10 @@ class L3HATestCase(framework.L3AgentTestFramework):
 
         # Add a second IPv6 subnet to the router internal interface.
         self._add_internal_interface_by_subnet(router.router, count=1,
-                ip_version=6, ipv6_subnet_modes=[slaac_mode],
+                ip_version=constants.IP_VERSION_6,
+                ipv6_subnet_modes=[slaac_mode],
                 interface_id=interface_id)
-        router.process(self.agent)
+        router.process()
         common_utils.wait_until_true(lambda: router.ha_state == 'master')
 
         # Verify that router internal interface is present and is configured
@@ -239,7 +282,7 @@ class L3HATestCase(framework.L3AgentTestFramework):
         subnets.append(interfaces[0]['subnets'][0])
         interfaces[0].update({'fixed_ips': fixed_ips, 'subnets': subnets})
         router.router[constants.INTERFACE_KEY] = interfaces
-        router.process(self.agent)
+        router.process()
 
         # Verify that router internal interface has a single ipaddress
         internal_iface = router.router[constants.INTERFACE_KEY][0]
@@ -268,7 +311,7 @@ class L3HATestCase(framework.L3AgentTestFramework):
         interface_name = router.get_external_device_interface_name(ex_gw_port)
         common_utils.wait_until_true(lambda: router.ha_state == 'master')
         self._add_fip(router, '172.168.1.20', fixed_address='10.0.0.3')
-        router.process(self.agent)
+        router.process()
         router.router[constants.FLOATINGIP_KEY] = []
         # The purpose of the test is to simply make sure no exception is raised
         # Because router.process will consume the FloatingIpSetupException,
@@ -287,11 +330,64 @@ class L3HATestCase(framework.L3AgentTestFramework):
         self.agent._process_updated_router(router1.router)
         common_utils.wait_until_true(lambda: router1.ha_state == 'master')
 
+    def test_ha_router_namespace_has_ip_nonlocal_bind_disabled(self):
+        router_info = self.generate_router_info(enable_ha=True)
+        router = self.manage_router(self.agent, router_info)
+        try:
+            ip_nonlocal_bind_value = ip_lib.get_ip_nonlocal_bind(
+                router.router_namespace.name)
+        except RuntimeError as rte:
+            stat_message = 'cannot stat /proc/sys/net/ipv4/ip_nonlocal_bind'
+            if stat_message in str(rte):
+                raise self.skipException(
+                    "This kernel doesn't support %s in network namespaces." % (
+                        ip_lib.IP_NONLOCAL_BIND))
+            raise
+        self.assertEqual(0, ip_nonlocal_bind_value)
+
+    @testtools.skipUnless(ipv6_utils.is_enabled_and_bind_by_default(),
+                          "IPv6 is not enabled")
+    def test_ha_router_namespace_has_ipv6_forwarding_disabled(self):
+        router_info = self.generate_router_info(enable_ha=True)
+        router_info[constants.HA_INTERFACE_KEY]['status'] = (
+            constants.PORT_STATUS_DOWN)
+        router = self.manage_router(self.agent, router_info)
+        external_port = router.get_ex_gw_port()
+        external_device_name = router.get_external_device_name(
+            external_port['id'])
+
+        common_utils.wait_until_true(lambda: router.ha_state == 'backup')
+        self._wait_until_ipv6_forwarding_has_state(
+            router.ns_name, external_device_name, 0)
+
+        router.router[constants.HA_INTERFACE_KEY]['status'] = (
+            constants.PORT_STATUS_ACTIVE)
+        self.agent._process_updated_router(router.router)
+        common_utils.wait_until_true(lambda: router.ha_state == 'master')
+        self._wait_until_ipv6_forwarding_has_state(
+            router.ns_name, external_device_name, 1)
+
+    @testtools.skipUnless(ipv6_utils.is_enabled_and_bind_by_default(),
+                          "IPv6 is not enabled")
+    def test_ha_router_without_gw_ipv6_forwarding_state(self):
+        router_info = self.generate_router_info(
+            enable_ha=True, enable_gw=False)
+        router_info[constants.HA_INTERFACE_KEY]['status'] = (
+            constants.PORT_STATUS_DOWN)
+        router = self.manage_router(self.agent, router_info)
+
+        common_utils.wait_until_true(lambda: router.ha_state == 'backup')
+        self._wait_until_ipv6_forwarding_has_state(router.ns_name, 'all', 0)
+
+        router.router[constants.HA_INTERFACE_KEY]['status'] = (
+            constants.PORT_STATUS_ACTIVE)
+        self.agent._process_updated_router(router.router)
+        common_utils.wait_until_true(lambda: router.ha_state == 'master')
+        self._wait_until_ipv6_forwarding_has_state(router.ns_name, 'all', 1)
+
 
 class L3HATestFailover(framework.L3AgentTestFramework):
 
-    NESTED_NAMESPACE_SEPARATOR = '@'
-
     def setUp(self):
         super(L3HATestFailover, self).setUp()
         conf = self._configure_agent('agent2')
@@ -302,35 +398,107 @@ class L3HATestFailover(framework.L3AgentTestFramework):
         br_int_2 = self._get_agent_ovs_integration_bridge(self.failover_agent)
 
         veth1, veth2 = self.useFixture(net_helpers.VethFixture()).ports
+        veth1.link.set_up()
+        veth2.link.set_up()
         br_int_1.add_port(veth1.name)
         br_int_2.add_port(veth2.name)
 
+    @staticmethod
+    def fail_gw_router_port(router):
+        # NOTE(slaweq): in HA failover tests there are two integration bridges
+        # connected with veth pair to each other. To stop traffic from router's
+        # namespace to gw ip (19.4.4.1) it needs to be blocked by openflow rule
+        # as simple setting ovs_integration_bridge device DOWN will not be
+        # enough because same IP address is also configured on
+        # ovs_integration_bridge device from second router and it will still
+        # respond to ping
+        r_br = ovs_lib.OVSBridge(router.driver.conf.ovs_integration_bridge)
+        external_port = router.get_ex_gw_port()
+        for subnet in external_port['subnets']:
+            r_br.add_flow(
+                proto='ip', nw_dst=subnet['gateway_ip'], actions='drop')
+
+    @staticmethod
+    def restore_gw_router_port(router):
+        r_br = ovs_lib.OVSBridge(router.driver.conf.ovs_integration_bridge)
+        external_port = router.get_ex_gw_port()
+        for subnet in external_port['subnets']:
+            r_br.delete_flows(proto='ip', nw_dst=subnet['gateway_ip'])
+
     def test_ha_router_failover(self):
-        router_info = self.generate_router_info(enable_ha=True)
-        get_ns_name = mock.patch.object(
-            namespaces.RouterNamespace, '_get_ns_name').start()
-        get_ns_name.return_value = "%s%s%s" % (
-            'qrouter-' + router_info['id'],
-            self.NESTED_NAMESPACE_SEPARATOR, self.agent.host)
-        router1 = self.manage_router(self.agent, router_info)
+        router1, router2 = self.create_ha_routers()
 
-        router_info_2 = copy.deepcopy(router_info)
-        router_info_2[constants.HA_INTERFACE_KEY] = (
-            l3_test_common.get_ha_interface(ip='169.254.192.2',
-                                            mac='22:22:22:22:22:22'))
+        master_router, slave_router = self._get_master_and_slave_routers(
+            router1, router2)
 
-        get_ns_name.return_value = "%s%s%s" % (
-            namespaces.RouterNamespace._get_ns_name(router_info_2['id']),
-            self.NESTED_NAMESPACE_SEPARATOR, self.failover_agent.host)
-        router2 = self.manage_router(self.failover_agent, router_info_2)
-
-        common_utils.wait_until_true(lambda: router1.ha_state == 'master')
-        common_utils.wait_until_true(lambda: router2.ha_state == 'backup')
+        self._assert_ipv6_accept_ra(master_router, True)
+        self._assert_ipv6_forwarding(master_router, True, True)
+        self._assert_ipv6_accept_ra(slave_router, False)
+        self._assert_ipv6_forwarding(slave_router, False, False)
 
         self.fail_ha_router(router1)
 
-        common_utils.wait_until_true(lambda: router2.ha_state == 'master')
-        common_utils.wait_until_true(lambda: router1.ha_state == 'backup')
+        # NOTE: passing slave_router as first argument, because we expect
+        # that this router should be the master
+        new_master, new_slave = self._get_master_and_slave_routers(
+            slave_router, master_router)
+
+        self.assertEqual(master_router, new_slave)
+        self.assertEqual(slave_router, new_master)
+        self._assert_ipv6_accept_ra(new_master, True)
+        self._assert_ipv6_forwarding(new_master, True, True)
+        self._assert_ipv6_accept_ra(new_slave, False)
+        # after transition from master -> slave, 'all' IPv6 forwarding should
+        # be enabled
+        self._assert_ipv6_forwarding(new_slave, False, True)
+
+    def test_ha_router_lost_gw_connection(self):
+        self.agent.conf.set_override(
+            'ha_vrrp_health_check_interval', 5)
+        self.failover_agent.conf.set_override(
+            'ha_vrrp_health_check_interval', 5)
+
+        router1, router2 = self.create_ha_routers()
+
+        master_router, slave_router = self._get_master_and_slave_routers(
+            router1, router2)
+
+        self.fail_gw_router_port(master_router)
+
+        # NOTE: passing slave_router as first argument, because we expect
+        # that this router should be the master
+        new_master, new_slave = self._get_master_and_slave_routers(
+            slave_router, master_router)
+
+        self.assertEqual(master_router, new_slave)
+        self.assertEqual(slave_router, new_master)
+
+    def test_both_ha_router_lost_gw_connection(self):
+        self.agent.conf.set_override(
+            'ha_vrrp_health_check_interval', 5)
+        self.failover_agent.conf.set_override(
+            'ha_vrrp_health_check_interval', 5)
+
+        router1, router2 = self.create_ha_routers()
+
+        master_router, slave_router = self._get_master_and_slave_routers(
+            router1, router2)
+
+        self.fail_gw_router_port(master_router)
+        self.fail_gw_router_port(slave_router)
+
+        common_utils.wait_until_true(
+            lambda: master_router.ha_state == 'master')
+        common_utils.wait_until_true(
+            lambda: slave_router.ha_state == 'master')
+
+        self.restore_gw_router_port(master_router)
+
+        new_master, new_slave = self._get_master_and_slave_routers(
+            master_router, slave_router)
+
+        self.assertEqual(master_router, new_master)
+        self.assertEqual(slave_router, new_slave)
 
 
 class LinuxBridgeL3HATestCase(L3HATestCase):
