@@ -20,6 +20,7 @@ from oslo_log import log
 from oslo_utils import uuidutils
 
 from neutron.agent.l3 import agent as l3_agent
+from neutron.agent.l3 import dvr_edge_ha_router as dvr_edge_ha_rtr
 from neutron.agent.l3 import dvr_edge_router as dvr_edge_rtr
 from neutron.agent.l3 import dvr_local_router as dvr_router
 from neutron.agent.l3 import link_local_allocator as lla
@@ -99,9 +100,8 @@ class TestDvrRouterOperations(base.BaseTestCase):
         self.mock_ip = mock.MagicMock()
         ip_cls.return_value = self.mock_ip
 
-        ip_rule = mock.patch('neutron.agent.linux.ip_lib.IPRule').start()
-        self.mock_rule = mock.MagicMock()
-        ip_rule.return_value = self.mock_rule
+        self.mock_delete_ip_rule = mock.patch.object(ip_lib,
+                                                     'delete_ip_rule').start()
 
         ip_dev = mock.patch('neutron.agent.linux.ip_lib.IPDevice').start()
         self.mock_ip_dev = mock.MagicMock()
@@ -205,6 +205,7 @@ class TestDvrRouterOperations(base.BaseTestCase):
         ri.rtr_fip_connect = False
         ex_gw_port = {'network_id': 'fake_net_id'}
         ri.create_dvr_external_gateway_on_agent(ex_gw_port)
+        ri._check_rtr_2_fip_connect = mock.Mock()
         ri.connect_rtr_2_fip()
         self.assertTrue(ri._check_if_address_scopes_match.called)
         if address_scopes_match:
@@ -259,9 +260,11 @@ class TestDvrRouterOperations(base.BaseTestCase):
         dnat_from_floatingip_to_fixedip = (
             'PREROUTING', '-d %s/32 -i %s -j DNAT --to-destination %s' % (
                 floating_ip, rtr_2_fip_name, fixed_ip))
-        snat_from_fixedip_to_floatingip = (
-            'float-snat', '-s %s/32 -j SNAT --to-source %s' % (
-                fixed_ip, floating_ip))
+        to_source = '-s %s/32 -j SNAT --to-source %s' % (fixed_ip, floating_ip)
+
+        if ri.iptables_manager.random_fully:
+            to_source += ' --random-fully'
+        snat_from_fixedip_to_floatingip = ('float-snat', to_source)
         actual = ri.floating_forward_rules(fip)
         expected = [dnat_from_floatingip_to_fixedip,
                     snat_from_fixedip_to_floatingip]
@@ -302,8 +305,9 @@ class TestDvrRouterOperations(base.BaseTestCase):
 
     @mock.patch.object(ip_lib, 'send_ip_addr_adv_notif')
     @mock.patch.object(ip_lib, 'IPDevice')
-    @mock.patch.object(ip_lib, 'IPRule')
-    def test_floating_ip_added_dist(self, mIPRule, mIPDevice, mock_adv_notif):
+    @mock.patch.object(ip_lib, 'add_ip_rule')
+    def test_floating_ip_added_dist(self, mock_add_ip_rule, mIPDevice,
+                                    mock_adv_notif):
         router = mock.MagicMock()
         ri = self._create_router(router)
         ri.ex_gw_port = ri.router['gw_port']
@@ -328,6 +332,7 @@ class TestDvrRouterOperations(base.BaseTestCase):
         ri.fip_ns = mock.Mock()
         ri.fip_ns.agent_gateway_port = agent_gw_port
         ri.create_dvr_external_gateway_on_agent(ri.ex_gw_port)
+        ri._check_rtr_2_fip_connect = mock.Mock()
         ri.connect_rtr_2_fip()
         self.assertTrue(ri.rtr_fip_connect)
         ri.fip_ns.allocate_rule_priority.return_value = FIP_PRI
@@ -337,25 +342,24 @@ class TestDvrRouterOperations(base.BaseTestCase):
         ri.fip_ns.local_subnets.allocate.return_value = subnet
         ip_cidr = common_utils.ip_to_cidr(fip['floating_ip_address'])
         ri.floating_ip_added_dist(fip, ip_cidr)
-        mIPRule().rule.add.assert_called_with(ip='192.168.0.1',
-                                              table=16,
-                                              priority=FIP_PRI)
+        mock_add_ip_rule.assert_called_with(
+            namespace=ri.router_namespace.name, ip='192.168.0.1',
+            table=16, priority=FIP_PRI)
         ri.fip_ns.local_subnets.allocate.assert_not_called()
 
         # Validate that fip_ns.local_subnets is called when
         # ri.rtr_fip_subnet is None
         ri.rtr_fip_subnet = None
         ri.floating_ip_added_dist(fip, ip_cidr)
-        mIPRule().rule.add.assert_called_with(ip='192.168.0.1',
-                                              table=16,
-                                              priority=FIP_PRI)
+        mock_add_ip_rule.assert_called_with(
+            namespace=ri.router_namespace.name, ip='192.168.0.1',
+            table=16, priority=FIP_PRI)
         ri.fip_ns.local_subnets.allocate.assert_called_once_with(ri.router_id)
         # TODO(mrsmith): add more asserts
 
     @mock.patch.object(ip_lib, 'IPWrapper')
     @mock.patch.object(ip_lib, 'IPDevice')
-    @mock.patch.object(ip_lib, 'IPRule')
-    def test_floating_ip_removed_dist(self, mIPRule, mIPDevice, mIPWrapper):
+    def test_floating_ip_removed_dist(self, mIPDevice, mIPWrapper):
         router = mock.MagicMock()
         ri = self._create_router(router)
         ri.ex_gw_port = ri.router['gw_port']
@@ -381,13 +385,13 @@ class TestDvrRouterOperations(base.BaseTestCase):
         ri.rtr_fip_subnet = s
         ri.fip_ns.local_subnets = mock.Mock()
         ri.floating_ip_removed_dist(fip_cidr)
-        mIPRule().rule.delete.assert_called_with(
-            ip=fixed_ip, table=16, priority=FIP_PRI)
+        self.mock_delete_ip_rule.assert_called_with(
+            ri.router_namespace.name, ip=fixed_ip, table=16, priority=FIP_PRI)
         mIPDevice().route.delete_route.assert_called_with(fip_cidr, str(s.ip))
         ri.fip_ns.local_subnets.allocate.assert_not_called()
 
-    @mock.patch.object(ip_lib, 'IPRule')
-    def test_floating_ip_moved_dist(self, mIPRule):
+    @mock.patch.object(ip_lib, 'add_ip_rule')
+    def test_floating_ip_moved_dist(self, mock_add_ip_rule):
         router = mock.MagicMock()
         ri = self._create_router(router)
         floating_ip_address = '15.1.2.3'
@@ -399,15 +403,15 @@ class TestDvrRouterOperations(base.BaseTestCase):
         ri.fip_ns.allocate_rule_priority.return_value = FIP_PRI
         ri.floating_ip_moved_dist(fip)
 
-        mIPRule().rule.delete.assert_called_once_with(
-            ip=fixed_ip, table=16, priority=FIP_PRI)
+        self.mock_delete_ip_rule.assert_called_once_with(
+            ri.router_namespace.name, ip=fixed_ip, table=16, priority=FIP_PRI)
         ri.fip_ns.deallocate_rule_priority.assert_called_once_with(
             floating_ip_address)
         ri.fip_ns.allocate_rule_priority.assert_called_once_with(
             floating_ip_address)
-        mIPRule().rule.add.assert_called_with(ip=fixed_ip,
-                                              table=16,
-                                              priority=FIP_PRI)
+        mock_add_ip_rule.assert_called_with(
+            namespace=ri.router_namespace.name, ip=fixed_ip,
+            table=16, priority=FIP_PRI)
 
     def _test_add_floating_ip(self, ri, fip, is_failure=False):
         if not is_failure:
@@ -521,49 +525,7 @@ class TestDvrRouterOperations(base.BaseTestCase):
             ri._set_subnet_arp_info(subnet_id)
         self.assertEqual(1, parp.call_count)
         self.mock_ip_dev.neigh.add.assert_called_once_with(
-            '1.2.3.4', '00:11:22:33:44:55', nud_state='permanent')
-
-        # Test negative case
-        router['distributed'] = False
-        ri._set_subnet_arp_info(subnet_id)
-        self.mock_ip_dev.neigh.add.never_called()
-
-    def test__set_subnet_arp_info_with_allowed_address_pair_port(self):
-        agent = l3_agent.L3NATAgent(HOSTNAME, self.conf)
-        router = l3_test_common.prepare_router_data(num_internal_ports=2)
-        router['distributed'] = True
-        self._set_ri_kwargs(agent, router['id'], router)
-        ri = dvr_router.DvrLocalRouter(HOSTNAME, **self.ri_kwargs)
-        ports = ri.router.get(lib_constants.INTERFACE_KEY, [])
-        subnet_id = l3_test_common.get_subnet_id(ports[0])
-        test_ports = [{'mac_address': '00:11:22:33:44:55',
-                       'device_owner': '',
-                       'fixed_ips': [{'ip_address': '1.2.3.4',
-                                      'prefixlen': 24,
-                                      'subnet_id': subnet_id}]},
-                      {'mac_address': '11:22:33:44:55:66',
-                       'device_owner': lib_constants.DEVICE_OWNER_LOADBALANCER,
-                       'fixed_ips': [{'ip_address': '1.2.3.5',
-                                      'prefixlen': 24,
-                                      'subnet_id': subnet_id}]},
-                      {'mac_address': '22:33:44:55:66:77',
-                       'device_owner':
-                           lib_constants.DEVICE_OWNER_LOADBALANCERV2,
-                       'fixed_ips': [{'ip_address': '1.2.3.6',
-                                      'prefixlen': 24,
-                                      'subnet_id': subnet_id}]}]
-
-        self.plugin_api.get_ports_by_subnet.return_value = test_ports
-
-        # Test basic case
-        ports[0]['subnets'] = [{'id': subnet_id,
-                                'cidr': '1.2.3.0/24'}]
-        with mock.patch.object(ri,
-                               '_process_arp_cache_for_internal_port') as parp:
-            ri._set_subnet_arp_info(subnet_id)
-        self.assertEqual(1, parp.call_count)
-        self.mock_ip_dev.neigh.add.assert_called_once_with(
-            '1.2.3.4', '00:11:22:33:44:55', nud_state='reachable')
+            '1.2.3.4', '00:11:22:33:44:55')
 
         # Test negative case
         router['distributed'] = False
@@ -578,33 +540,14 @@ class TestDvrRouterOperations(base.BaseTestCase):
             router[lib_constants.INTERFACE_KEY][0])
         arp_table = {'ip_address': '1.7.23.11',
                      'mac_address': '00:11:22:33:44:55',
-                     'subnet_id': subnet_id,
-                     'nud_state': 'permanent'}
+                     'subnet_id': subnet_id}
 
         payload = {'arp_table': arp_table, 'router_id': router['id']}
         agent._router_added(router['id'], router)
         agent.add_arp_entry(None, payload)
         agent.router_deleted(None, router['id'])
         self.mock_ip_dev.neigh.add.assert_called_once_with(
-            '1.7.23.11', '00:11:22:33:44:55', nud_state='permanent')
-
-    def test_add_arp_entry_with_nud_state_reachable(self):
-        agent = l3_agent.L3NATAgent(HOSTNAME, self.conf)
-        router = l3_test_common.prepare_router_data(num_internal_ports=2)
-        router['distributed'] = True
-        subnet_id = l3_test_common.get_subnet_id(
-            router[lib_constants.INTERFACE_KEY][0])
-        arp_table = {'ip_address': '1.7.23.11',
-                     'mac_address': '00:11:22:33:44:55',
-                     'subnet_id': subnet_id,
-                     'nud_state': 'reachable'}
-
-        payload = {'arp_table': arp_table, 'router_id': router['id']}
-        agent._router_added(router['id'], router)
-        agent.add_arp_entry(None, payload)
-        agent.router_deleted(None, router['id'])
-        self.mock_ip_dev.neigh.add.assert_called_once_with(
-            '1.7.23.11', '00:11:22:33:44:55', nud_state='reachable')
+            '1.7.23.11', '00:11:22:33:44:55')
 
     def test_add_arp_entry_no_routerinfo(self):
         agent = l3_agent.L3NATAgent(HOSTNAME, self.conf)
@@ -623,9 +566,9 @@ class TestDvrRouterOperations(base.BaseTestCase):
                             'foo_router_id',
                             {'distributed': True, 'gw_port_host': HOSTNAME})
         ri = dvr_router.DvrLocalRouter(HOSTNAME, **self.ri_kwargs)
-        with mock.patch.object(l3_agent.ip_lib, 'IPDevice') as f:
-            ri._update_arp_entry(mock.ANY, mock.ANY, 'foo_subnet_id', 'add')
-        self.assertFalse(f.call_count)
+        ri.get_internal_device_name = mock.Mock()
+        ri._update_arp_entry(mock.ANY, mock.ANY, 'foo_subnet_id', 'add')
+        self.assertFalse(ri.get_internal_device_name.call_count)
 
     def _setup_test_for_arp_entry_cache(self):
         agent = l3_agent.L3NATAgent(HOSTNAME, self.conf)
@@ -640,11 +583,11 @@ class TestDvrRouterOperations(base.BaseTestCase):
     def test__update_arp_entry_calls_arp_cache_with_no_device(self):
         ri, subnet_id = self._setup_test_for_arp_entry_cache()
         state = True
-        with mock.patch.object(l3_agent.ip_lib, 'IPDevice') as rtrdev,\
+        with mock.patch('neutron.agent.linux.ip_lib.IPDevice') as rtrdev,\
                 mock.patch.object(ri, '_cache_arp_entry') as arp_cache:
-                rtrdev.return_value.exists.return_value = False
-                state = ri._update_arp_entry(
-                    mock.ANY, mock.ANY, subnet_id, 'add')
+            rtrdev.return_value.exists.return_value = False
+            state = ri._update_arp_entry(
+                mock.ANY, mock.ANY, subnet_id, 'add')
         self.assertFalse(state)
         self.assertTrue(arp_cache.called)
         arp_cache.assert_called_once_with(mock.ANY, mock.ANY,
@@ -794,6 +737,9 @@ class TestDvrRouterOperations(base.BaseTestCase):
         agent._fetch_external_net_id = mock.Mock(return_value=external_net_id)
         ri.ex_gw_port = ri.router['gw_port']
         del ri.router['gw_port']
+        ri.external_gateway_added(
+            ri.ex_gw_port,
+            ri.get_external_device_name(ri.ex_gw_port['id']))
         ri.fip_ns = None
         nat = ri.iptables_manager.ipv4['nat']
         nat.clear_rules_by_tag = mock.Mock()
@@ -869,3 +815,76 @@ class TestDvrRouterOperations(base.BaseTestCase):
                                                                         ri)
         ri._handle_router_snat_rules(ex_gw_port, interface_name)
         ipv4_nat.add_rule.assert_called_once_with('snat', '-j $float-snat')
+
+    @mock.patch.object(dvr_edge_rtr.DvrEdgeRouter,
+                       'add_centralized_floatingip')
+    def test_add_centralized_floatingip_dvr_ha(
+            self,
+            super_add_centralized_floatingip):
+        agent = l3_agent.L3NATAgent(HOSTNAME, self.conf)
+        agent.conf.agent_mode = lib_constants.L3_AGENT_MODE_DVR_SNAT
+        router = l3_test_common.prepare_router_data(
+            num_internal_ports=2, enable_ha=True)
+        router['gw_port_host'] = HOSTNAME
+        self.mock_driver.unplug.reset_mock()
+        self._set_ri_kwargs(agent, router['id'], router)
+        fip = {'id': _uuid()}
+        fip_cidr = '11.22.33.44/24'
+
+        ri = dvr_edge_ha_rtr.DvrEdgeHaRouter(HOSTNAME, [], **self.ri_kwargs)
+        ri.is_router_master = mock.Mock(return_value=False)
+        ri._add_vip = mock.Mock()
+        interface_name = ri.get_snat_external_device_interface_name(
+            ri.get_ex_gw_port())
+        ri.add_centralized_floatingip(fip, fip_cidr)
+        ri._add_vip.assert_called_once_with(fip_cidr, interface_name)
+        super_add_centralized_floatingip.assert_not_called()
+
+        router[lib_constants.HA_INTERFACE_KEY]['status'] = 'DOWN'
+        self._set_ri_kwargs(agent, router['id'], router)
+        ri_1 = dvr_edge_ha_rtr.DvrEdgeHaRouter(HOSTNAME, [], **self.ri_kwargs)
+        ri_1.is_router_master = mock.Mock(return_value=True)
+        ri_1._add_vip = mock.Mock()
+        interface_name = ri_1.get_snat_external_device_interface_name(
+            ri_1.get_ex_gw_port())
+        ri_1.add_centralized_floatingip(fip, fip_cidr)
+        ri_1._add_vip.assert_called_once_with(fip_cidr, interface_name)
+        super_add_centralized_floatingip.assert_not_called()
+
+        router[lib_constants.HA_INTERFACE_KEY]['status'] = 'ACTIVE'
+        self._set_ri_kwargs(agent, router['id'], router)
+        ri_2 = dvr_edge_ha_rtr.DvrEdgeHaRouter(HOSTNAME, [], **self.ri_kwargs)
+        ri_2.is_router_master = mock.Mock(return_value=True)
+        ri_2._add_vip = mock.Mock()
+        interface_name = ri_2.get_snat_external_device_interface_name(
+            ri_2.get_ex_gw_port())
+        ri_2.add_centralized_floatingip(fip, fip_cidr)
+        ri_2._add_vip.assert_called_once_with(fip_cidr, interface_name)
+        super_add_centralized_floatingip.assert_called_once_with(fip,
+                                                                 fip_cidr)
+
+    @mock.patch.object(dvr_edge_rtr.DvrEdgeRouter,
+                       'remove_centralized_floatingip')
+    def test_remove_centralized_floatingip(self,
+                                super_remove_centralized_floatingip):
+        agent = l3_agent.L3NATAgent(HOSTNAME, self.conf)
+        agent.conf.agent_mode = lib_constants.L3_AGENT_MODE_DVR_SNAT
+        router = l3_test_common.prepare_router_data(num_internal_ports=2)
+        router['gw_port_host'] = HOSTNAME
+        self.mock_driver.unplug.reset_mock()
+        self._set_ri_kwargs(agent, router['id'], router)
+        fip_cidr = '11.22.33.44/24'
+
+        ri = dvr_edge_ha_rtr.DvrEdgeHaRouter(HOSTNAME, [], **self.ri_kwargs)
+        ri.is_router_master = mock.Mock(return_value=False)
+        ri._remove_vip = mock.Mock()
+        ri.remove_centralized_floatingip(fip_cidr)
+        ri._remove_vip.assert_called_once_with(fip_cidr)
+        super_remove_centralized_floatingip.assert_not_called()
+
+        ri1 = dvr_edge_ha_rtr.DvrEdgeHaRouter(HOSTNAME, [], **self.ri_kwargs)
+        ri1.is_router_master = mock.Mock(return_value=True)
+        ri1._remove_vip = mock.Mock()
+        ri1.remove_centralized_floatingip(fip_cidr)
+        ri1._remove_vip.assert_called_once_with(fip_cidr)
+        super_remove_centralized_floatingip.assert_called_once_with(fip_cidr)
