@@ -12,17 +12,22 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+import copy
+import signal
 import socket
 
 import mock
 import six
 import testtools
 
+from neutron_lib import exceptions
+from oslo_config import cfg
 import oslo_i18n
 
 from neutron.agent.linux import utils
 from neutron.tests import base
 from neutron.tests.common import helpers
+from neutron.tests import tools
 
 
 _marker = object()
@@ -37,6 +42,16 @@ class AgentUtilsExecuteTest(base.BaseTestCase):
         self.process.return_value.returncode = 0
         self.mock_popen = self.process.return_value.communicate
 
+    def test_xenapi_root_helper(self):
+        token = utils.xenapi_root_helper.ROOT_HELPER_DAEMON_TOKEN
+        self.config(group='AGENT', root_helper_daemon=token)
+        with mock.patch(
+                'neutron.agent.linux.utils.xenapi_root_helper.XenAPIClient')\
+                as mock_xenapi_class:
+            mock_client = mock_xenapi_class.return_value
+            cmd_client = utils.RootwrapDaemonHelper.get_client()
+            self.assertEqual(cmd_client, mock_client)
+
     def test_without_helper(self):
         expected = "%s\n" % self.test_file
         self.mock_popen.return_value = [expected, ""]
@@ -50,6 +65,17 @@ class AgentUtilsExecuteTest(base.BaseTestCase):
         result = utils.execute(["ls", self.test_file], run_as_root=True)
         self.assertEqual(result, expected)
 
+    @mock.patch.object(utils.RootwrapDaemonHelper, 'get_client')
+    def test_with_helper_exception(self, get_client):
+        client_inst = mock.Mock()
+        client_inst.execute.side_effect = RuntimeError
+        get_client.return_value = client_inst
+        self.config(group='AGENT', root_helper_daemon='echo')
+        with mock.patch.object(utils, 'LOG') as log:
+            self.assertRaises(RuntimeError, utils.execute,
+                              ['ls'], run_as_root=True)
+            self.assertTrue(log.error.called)
+
     def test_stderr_true(self):
         expected = "%s\n" % self.test_file
         self.mock_popen.return_value = [expected, ""]
@@ -122,7 +148,7 @@ class AgentUtilsExecuteTest(base.BaseTestCase):
         self.mock_popen.return_value = ('', '')
         self.process.return_value.returncode = 1
         with mock.patch.object(utils, 'LOG') as log:
-            self.assertRaises(utils.ProcessExecutionError, utils.execute,
+            self.assertRaises(exceptions.ProcessExecutionError, utils.execute,
                               ['ls'], log_fail_as_error=False)
             self.assertFalse(log.error.called)
 
@@ -170,22 +196,147 @@ class AgentUtilsExecuteEncodeTest(base.BaseTestCase):
         self.assertEqual((str_data, ''), result)
 
 
-class AgentUtilsGetInterfaceMAC(base.BaseTestCase):
-    def test_get_interface_mac(self):
-        expect_val = '01:02:03:04:05:06'
-        with mock.patch('fcntl.ioctl') as ioctl:
-            ioctl.return_value = ''.join(['\x00' * 18,
-                                          '\x01\x02\x03\x04\x05\x06',
-                                          '\x00' * 232])
-            actual_val = utils.get_interface_mac('eth0')
-        self.assertEqual(actual_val, expect_val)
+class TestFindParentPid(base.BaseTestCase):
+    def setUp(self):
+        super(TestFindParentPid, self).setUp()
+        self.m_execute = mock.patch.object(utils, 'execute').start()
+
+    def test_returns_none_for_no_valid_pid(self):
+        self.m_execute.side_effect = exceptions.ProcessExecutionError(
+            '', returncode=1)
+        self.assertIsNone(utils.find_parent_pid(-1))
+
+    def test_returns_parent_id_for_good_ouput(self):
+        self.m_execute.return_value = '123 \n'
+        self.assertEqual(utils.find_parent_pid(-1), '123')
+
+    def test_raises_exception_returncode_0(self):
+        with testtools.ExpectedException(exceptions.ProcessExecutionError):
+            self.m_execute.side_effect = \
+                exceptions.ProcessExecutionError('', returncode=0)
+            utils.find_parent_pid(-1)
+
+    def test_raises_unknown_exception(self):
+        with testtools.ExpectedException(RuntimeError):
+            self.m_execute.side_effect = RuntimeError()
+            utils.find_parent_pid(-1)
+
+
+class TestFindForkTopParent(base.BaseTestCase):
+    def _test_find_fork_top_parent(self, expected=_marker,
+                                   find_parent_pid_retvals=None,
+                                   pid_invoked_with_cmdline_retvals=None):
+        def _find_parent_pid(x):
+            if find_parent_pid_retvals:
+                return find_parent_pid_retvals.pop(0)
+
+        pid_invoked_with_cmdline = {}
+        if pid_invoked_with_cmdline_retvals:
+            pid_invoked_with_cmdline['side_effect'] = (
+                pid_invoked_with_cmdline_retvals)
+        else:
+            pid_invoked_with_cmdline['return_value'] = False
+        with mock.patch.object(utils, 'find_parent_pid',
+                               side_effect=_find_parent_pid), \
+                mock.patch.object(utils, 'pid_invoked_with_cmdline',
+                                  **pid_invoked_with_cmdline):
+            actual = utils.find_fork_top_parent(_marker)
+        self.assertEqual(expected, actual)
+
+    def test_returns_own_pid_no_parent(self):
+        self._test_find_fork_top_parent()
+
+    def test_returns_own_pid_nofork(self):
+        self._test_find_fork_top_parent(find_parent_pid_retvals=['2', '3'])
+
+    def test_returns_first_parent_pid_fork(self):
+        self._test_find_fork_top_parent(
+            expected='2',
+            find_parent_pid_retvals=['2', '3', '4'],
+            pid_invoked_with_cmdline_retvals=[True, False, False])
+
+    def test_returns_top_parent_pid_fork(self):
+        self._test_find_fork_top_parent(
+            expected='4',
+            find_parent_pid_retvals=['2', '3', '4'],
+            pid_invoked_with_cmdline_retvals=[True, True, True])
+
+
+class TestKillProcess(base.BaseTestCase):
+    def _test_kill_process(self, pid, raise_exception=False,
+                           kill_signal=signal.SIGKILL, pid_killed=True):
+        if raise_exception:
+            exc = exceptions.ProcessExecutionError('', returncode=0)
+        else:
+            exc = None
+        with mock.patch.object(utils, 'execute',
+                               side_effect=exc) as mock_execute:
+            with mock.patch.object(utils, 'process_is_running',
+                                   return_value=not pid_killed):
+                utils.kill_process(pid, kill_signal, run_as_root=True)
+
+        mock_execute.assert_called_with(['kill', '-%d' % kill_signal, pid],
+                                        run_as_root=True)
+
+    def test_kill_process_returns_none_for_valid_pid(self):
+        self._test_kill_process('1')
+
+    def test_kill_process_returns_none_for_stale_pid(self):
+        self._test_kill_process('1', raise_exception=True)
+
+    def test_kill_process_raises_exception_for_execute_exception(self):
+        with testtools.ExpectedException(exceptions.ProcessExecutionError):
+            # Simulate that the process is running after trying to kill due to
+            # any reason such as, for example, Permission denied
+            self._test_kill_process('1', raise_exception=True,
+                                    pid_killed=False)
+
+    def test_kill_process_with_different_signal(self):
+        self._test_kill_process('1', kill_signal=signal.SIGTERM)
+
+
+class TestGetCmdlineFromPid(base.BaseTestCase):
+
+    def setUp(self):
+        super(TestGetCmdlineFromPid, self).setUp()
+        self.pid = 34
+        self.process_is_running_mock = mock.patch.object(
+            utils, "process_is_running").start()
+
+    def _test_cmdline(self, process, expected_cmd):
+        self.process_is_running_mock.return_value = True
+        mock_open = self.useFixture(
+            tools.OpenFixture('/proc/%s/cmdline' % self.pid, process)
+        ).mock_open
+        cmdline = utils.get_cmdline_from_pid(self.pid)
+        mock_open.assert_called_once_with('/proc/%s/cmdline' % self.pid, 'r')
+        self.assertEqual(expected_cmd, cmdline)
+
+    def test_cmdline_separated_with_null_char(self):
+        process_cmd = "python3\0test-binary\0test option\0"
+        expected_cmdline = ["python3", "test-binary", "test option"]
+        self._test_cmdline(process_cmd, expected_cmdline)
+
+    def test_cmdline_separated_with_space_char(self):
+        process_cmd = "python3 test-binary test option\0"
+        expected_cmdline = ["python3", "test-binary", "test", "option"]
+        self._test_cmdline(process_cmd, expected_cmdline)
+
+    def test_no_process_running(self):
+        self.process_is_running_mock.return_value = False
+        mock_open = self.useFixture(
+            tools.OpenFixture('/proc/%s/cmdline' % self.pid)
+        ).mock_open
+        cmdline = utils.get_cmdline_from_pid(self.pid)
+        mock_open.assert_not_called()
+        self.assertEqual([], cmdline)
 
 
 class TestFindChildPids(base.BaseTestCase):
 
     def test_returns_empty_list_for_exit_code_1(self):
         with mock.patch.object(utils, 'execute',
-                               side_effect=utils.ProcessExecutionError(
+                               side_effect=exceptions.ProcessExecutionError(
                                    '', returncode=1)):
             self.assertEqual([], utils.find_child_pids(-1))
 
@@ -197,6 +348,14 @@ class TestFindChildPids(base.BaseTestCase):
         with mock.patch.object(utils, 'execute', return_value=' 123 \n 185\n'):
             self.assertEqual(utils.find_child_pids(-1), ['123', '185'])
 
+    def test_returns_list_of_child_process_ids_recursively(self):
+        with mock.patch.object(utils, 'execute',
+                               side_effect=[' 123 \n 185\n',
+                                            ' 40 \n', '\n',
+                                            '41\n', '\n']):
+            actual = utils.find_child_pids(-1, True)
+            self.assertEqual(actual, ['123', '185', '40', '41'])
+
     def test_raises_unknown_exception(self):
         with testtools.ExpectedException(RuntimeError):
             with mock.patch.object(utils, 'execute',
@@ -222,10 +381,10 @@ class TestGetRoothelperChildPid(base.BaseTestCase):
             pid_invoked_with_cmdline['return_value'] = False
         with mock.patch.object(utils, 'find_child_pids',
                                side_effect=_find_child_pids), \
-            mock.patch.object(utils, 'pid_invoked_with_cmdline',
-                              **pid_invoked_with_cmdline):
-                actual = utils.get_root_helper_child_pid(
-                        mock_pid, mock.ANY, run_as_root)
+                mock.patch.object(utils, 'pid_invoked_with_cmdline',
+                                  **pid_invoked_with_cmdline):
+            actual = utils.get_root_helper_child_pid(
+                mock_pid, mock.ANY, run_as_root)
         if expected is _marker:
             expected = str(mock_pid)
         self.assertEqual(expected, actual)
@@ -364,13 +523,37 @@ class TestUnixDomainHttpConnection(base.BaseTestCase):
 
 
 class TestUnixDomainHttpProtocol(base.BaseTestCase):
+    def setUp(self):
+        super(TestUnixDomainHttpProtocol, self).setUp()
+        self.ewhi = mock.patch('eventlet.wsgi.HttpProtocol.__init__').start()
+
     def test_init_empty_client(self):
-        u = utils.UnixDomainHttpProtocol(mock.Mock(), '', mock.Mock())
-        self.assertEqual(u.client_address, ('<local>', 0))
+        for addr in ('', b''):
+            utils.UnixDomainHttpProtocol(mock.Mock(), addr, mock.Mock())
+            self.ewhi.assert_called_once_with(mock.ANY, mock.ANY,
+                                              ('<local>', 0), mock.ANY)
+            self.ewhi.reset_mock()
 
     def test_init_with_client(self):
-        u = utils.UnixDomainHttpProtocol(mock.Mock(), 'foo', mock.Mock())
-        self.assertEqual(u.client_address, 'foo')
+        utils.UnixDomainHttpProtocol(mock.Mock(), 'foo', mock.Mock())
+        self.ewhi.assert_called_once_with(mock.ANY, mock.ANY, 'foo', mock.ANY)
+
+    def test_init_new_style_empty_client(self):
+        conn_state = ['', mock.Mock(), mock.Mock()]
+        # have to make a copy since the init will modify what we pass
+        csc = copy.copy(conn_state)
+        csc[0] = ('<local>', 0)
+        utils.UnixDomainHttpProtocol(conn_state, mock.Mock())
+        self.ewhi.assert_called_once_with(mock.ANY, csc, mock.ANY)
+
+    def test_init_new_style_client(self):
+        conn_state = ['foo', mock.Mock(), mock.Mock()]
+        utils.UnixDomainHttpProtocol(conn_state, mock.Mock())
+        self.ewhi.assert_called_once_with(mock.ANY, conn_state, mock.ANY)
+
+    def test_init_unknown_client(self):
+        utils.UnixDomainHttpProtocol('foo')
+        self.ewhi.assert_called_once_with(mock.ANY, 'foo')
 
 
 class TestUnixDomainWSGIServer(base.BaseTestCase):
@@ -402,6 +585,7 @@ class TestUnixDomainWSGIServer(base.BaseTestCase):
             'app',
             protocol=utils.UnixDomainHttpProtocol,
             log=mock.ANY,
+            log_format=cfg.CONF.wsgi_log_format,
             max_size=self.server.num_threads
         )
 
@@ -416,5 +600,6 @@ class TestUnixDomainWSGIServer(base.BaseTestCase):
             'app',
             protocol=utils.UnixDomainHttpProtocol,
             log=mock.ANY,
+            log_format=cfg.CONF.wsgi_log_format,
             max_size=num_threads
         )
