@@ -14,9 +14,11 @@
 #    under the License.
 
 import copy
+
 import mock
 import netaddr
 from neutron_lib import constants
+from neutron_lib import context as ncontext
 from neutron_lib import exceptions as n_exc
 from oslo_config import cfg
 from oslo_db import exception as db_exc
@@ -24,11 +26,11 @@ from oslo_utils import netutils
 from oslo_utils import uuidutils
 import webob.exc
 
-from neutron.common import constants as n_const
 from neutron.db import ipam_backend_mixin
 from neutron.db import ipam_pluggable_backend
-from neutron.db import models_v2
 from neutron.ipam import requests as ipam_req
+from neutron.objects import ports as port_obj
+from neutron.objects import subnet as obj_subnet
 from neutron.tests.unit.db import test_db_base_plugin_v2 as test_db_base
 
 
@@ -65,6 +67,7 @@ class TestDbBasePluginIpam(test_db_base.NeutronDbPluginV2TestCase):
         super(TestDbBasePluginIpam, self).setUp()
         self.tenant_id = uuidutils.generate_uuid()
         self.subnet_id = uuidutils.generate_uuid()
+        self.admin_context = ncontext.get_admin_context()
 
     def _prepare_mocks(self, address_factory=None, subnet_factory=None):
         if address_factory is None:
@@ -90,7 +93,7 @@ class TestDbBasePluginIpam(test_db_base.NeutronDbPluginV2TestCase):
         mocks['driver'].allocate_subnet.return_value = mocks['subnet']
         mocks['driver'].get_allocator.return_value = mocks['subnets']
         mocks['subnets'].allocate.return_value = (
-            mock.sentinel.address, mock.sentinel.subnet_id)
+            '127.0.0.1', uuidutils.generate_uuid())
         mocks['driver'].get_subnet_request_factory.return_value = (
             subnet_factory)
         mocks['driver'].get_address_request_factory.return_value = (
@@ -112,12 +115,10 @@ class TestDbBasePluginIpam(test_db_base.NeutronDbPluginV2TestCase):
 
     def _get_allocate_mock(self, subnet_id, auto_ip='10.0.0.2',
                            fail_ip='127.0.0.1',
-                           exception=None):
-        if exception is None:
-            exception = n_exc.InvalidInput(error_message='SomeError')
-
+                           exception=n_exc.InvalidInput(
+                               error_message='SomeError')):
         def allocate_mock(request):
-            if type(request) == ipam_req.SpecificAddressRequest:
+            if isinstance(request, ipam_req.SpecificAddressRequest):
                 if request.address == netaddr.IPAddress(fail_ip):
                     raise exception
                 else:
@@ -127,15 +128,14 @@ class TestDbBasePluginIpam(test_db_base.NeutronDbPluginV2TestCase):
 
         return allocate_mock
 
-    def _get_deallocate_mock(self, fail_ip='127.0.0.1', exception=None):
-            if exception is None:
-                exception = n_exc.InvalidInput(error_message='SomeError')
+    def _get_deallocate_mock(self, fail_ip='127.0.0.1',
+                             exception=n_exc.InvalidInput(
+                                 error_message='SomeError')):
+        def deallocate_mock(ip):
+            if str(ip) == fail_ip:
+                raise exception
 
-            def deallocate_mock(ip):
-                if str(ip) == fail_ip:
-                    raise exception
-
-            return deallocate_mock
+        return deallocate_mock
 
     def _validate_allocate_calls(self, expected_calls, mocks):
         self.assertTrue(mocks['subnets'].allocate.called)
@@ -255,6 +255,34 @@ class TestDbBasePluginIpam(test_db_base.NeutronDbPluginV2TestCase):
         self.assertIsInstance(request, ipam_req.AutomaticAddressRequest)
         self.assertEqual(eui64_ip, request.address)
 
+    def test_allocate_multiple_eui64_ips(self):
+        mocks = self._prepare_ipam()
+        ips = [{'subnet_id': self._gen_subnet_id(),
+                'subnet_cidr': '2001:470:abcd::/64',
+                'mac': '6c:62:6d:de:cf:49',
+                'eui64_address': True},
+               {'subnet_id': self._gen_subnet_id(),
+                'subnet_cidr': '2001:360:abcd::/64',
+                'mac': '6c:62:6d:de:cf:49',
+                'eui64_address': True}]
+        mocks['ipam']._ipam_allocate_ips(mock.ANY, mocks['driver'],
+                                         mock.ANY, ips)
+
+        eui64_ips = []
+        request_ips = []
+        i = 0
+        requests = mocks['subnets'].allocate.call_args_list
+        for ip in ips:
+            eui64_ip = netutils.get_ipv6_addr_by_EUI64(ip['subnet_cidr'],
+                                                       ip['mac'])
+            self.assertIsInstance(requests[i][0][0],
+                                  ipam_req.AutomaticAddressRequest)
+            self.assertEqual(eui64_ip, requests[i][0][0].address)
+            request_ips.append(requests[i][0][0].address)
+            eui64_ips.append(eui64_ip)
+            i += 1
+        self.assertEqual(request_ips, eui64_ips)
+
     def test_allocate_multiple_ips(self):
         mocks = self._prepare_ipam()
         subnet_id = self._gen_subnet_id()
@@ -341,8 +369,8 @@ class TestDbBasePluginIpam(test_db_base.NeutronDbPluginV2TestCase):
     def test_test_fixed_ips_for_port_pd_gateway(self):
         context = mock.Mock()
         pluggable_backend = ipam_pluggable_backend.IpamPluggableBackend()
-        with self.subnet(cidr=n_const.PROVISIONAL_IPV6_PD_PREFIX,
-                         ip_version=6) as subnet:
+        with self.subnet(cidr=constants.PROVISIONAL_IPV6_PD_PREFIX,
+                         ip_version=constants.IP_VERSION_6) as subnet:
             subnet = subnet['subnet']
             fixed_ips = [{'subnet_id': subnet['id'],
                          'ip_address': '::1'}]
@@ -376,9 +404,9 @@ class TestDbBasePluginIpam(test_db_base.NeutronDbPluginV2TestCase):
     def test_create_ipv6_pd_subnet_over_ipam(self, pool_mock):
         mocks = self._prepare_mocks_with_pool_mock(pool_mock)
         cfg.CONF.set_override('ipv6_pd_enabled', True)
-        cidr = n_const.PROVISIONAL_IPV6_PD_PREFIX
+        cidr = constants.PROVISIONAL_IPV6_PD_PREFIX
         allocation_pools = [netaddr.IPRange('::2', '::ffff:ffff:ffff:ffff')]
-        with self.subnet(cidr=None, ip_version=6,
+        with self.subnet(cidr=None, ip_version=constants.IP_VERSION_6,
                          subnetpool_id=constants.IPV6_PD_POOL_ID,
                          ipv6_ra_mode=constants.IPV6_SLAAC,
                          ipv6_address_mode=constants.IPV6_SLAAC):
@@ -476,7 +504,7 @@ class TestDbBasePluginIpam(test_db_base.NeutronDbPluginV2TestCase):
                                    admin_state_up=True)
         network = self.deserialize(self.fmt, res)
         subnet = self._make_subnet(self.fmt, network, gateway_ip,
-                                   cidr, ip_version=4)
+                                   cidr, ip_version=constants.IP_VERSION_4)
         req = self.new_delete_request('subnets', subnet['subnet']['id'])
         res = req.get_response(self.api)
         self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
@@ -496,7 +524,7 @@ class TestDbBasePluginIpam(test_db_base.NeutronDbPluginV2TestCase):
                                    admin_state_up=True)
         network = self.deserialize(self.fmt, res)
         subnet = self._make_subnet(self.fmt, network, gateway_ip,
-                                   cidr, ip_version=4)
+                                   cidr, ip_version=constants.IP_VERSION_4)
         req = self.new_delete_request('subnets', subnet['subnet']['id'])
         res = req.get_response(self.api)
         self.assertEqual(webob.exc.HTTPServerError.code, res.status_int)
@@ -648,7 +676,7 @@ class TestDbBasePluginIpam(test_db_base.NeutronDbPluginV2TestCase):
         fixed_ips_mock = mock.Mock(return_value=changes.add)
         mocks['ipam'] = ipam_pluggable_backend.IpamPluggableBackend()
         mocks['ipam']._get_changed_ips_for_port = changes_mock
-        mocks['ipam']._ipam_get_subnets = mock.Mock()
+        mocks['ipam']._ipam_get_subnets = mock.Mock(return_value=[])
         mocks['ipam']._test_fixed_ips_for_port = fixed_ips_mock
         mocks['ipam']._update_ips_for_pd_subnet = mock.Mock(return_value=[])
 
@@ -659,7 +687,8 @@ class TestDbBasePluginIpam(test_db_base.NeutronDbPluginV2TestCase):
                                            original_ips, new_ips, mac)
         mocks['driver'].get_address_request_factory.assert_called_once_with()
         mocks['ipam']._ipam_get_subnets.assert_called_once_with(
-            context, network_id=port_dict['network_id'], host=None)
+            context, network_id=port_dict['network_id'], host=None,
+            service_type=port_dict['device_owner'])
         # Validate port_dict is passed into address_factory
         address_factory.get_request.assert_called_once_with(context,
                                                             port_dict,
@@ -667,7 +696,7 @@ class TestDbBasePluginIpam(test_db_base.NeutronDbPluginV2TestCase):
 
     @mock.patch('neutron.ipam.driver.Pool')
     def test_update_ips_for_port_passes_port_id_to_factory(self, pool_mock):
-        port_id = mock.Mock()
+        port_id = uuidutils.generate_uuid()
         network_id = uuidutils.generate_uuid()
         address_factory = mock.Mock()
         mocks = self._prepare_mocks_with_pool_mock(
@@ -682,7 +711,7 @@ class TestDbBasePluginIpam(test_db_base.NeutronDbPluginV2TestCase):
         subnets = [{'id': ip_dict['subnet_id'],
                     'network_id': network_id,
                     'cidr': '192.1.1.0/24',
-                    'ip_version': 4,
+                    'ip_version': constants.IP_VERSION_4,
                     'ipv6_address_mode': None,
                     'ipv6_ra_mode': None}]
         get_subnets_mock = mock.Mock(return_value=subnets)
@@ -691,15 +720,17 @@ class TestDbBasePluginIpam(test_db_base.NeutronDbPluginV2TestCase):
         mocks['ipam']._ipam_get_subnets = get_subnets_mock
         mocks['ipam']._get_subnet = get_subnet_mock
 
-        mocks['ipam'].allocate_ips_for_port_and_store(context,
-                                                      port_dict,
-                                                      port_id)
+        with mock.patch.object(port_obj.IPAllocation, 'create'):
+            mocks['ipam'].allocate_ips_for_port_and_store(context,
+                                                          port_dict,
+                                                          port_id)
 
         mocks['driver'].get_address_request_factory.assert_called_once_with()
 
         port_dict_with_id = port_dict['port'].copy()
         port_dict_with_id['id'] = port_id
         # Validate port id is added to port dict before address_factory call
+        ip_dict.pop('device_owner')
         address_factory.get_request.assert_called_once_with(context,
                                                             port_dict_with_id,
                                                             ip_dict)
@@ -709,13 +740,20 @@ class TestDbBasePluginIpam(test_db_base.NeutronDbPluginV2TestCase):
     def _test_update_db_subnet(self, pool_mock, subnet, expected_subnet,
                                old_pools):
         subnet_factory = mock.Mock()
-        context = mock.Mock()
+        context = self.admin_context
+
+        if 'cidr' in subnet:
+            subnet['cidr'] = netaddr.IPNetwork(subnet['cidr'])
+        if 'cidr' in expected_subnet:
+            expected_subnet['cidr'] = netaddr.IPNetwork(
+                expected_subnet['cidr'])
 
         mocks = self._prepare_mocks_with_pool_mock(
             pool_mock, subnet_factory=subnet_factory)
 
         mocks['ipam'] = ipam_pluggable_backend.IpamPluggableBackend()
-        mocks['ipam'].update_db_subnet(context, id, subnet, old_pools)
+        mocks['ipam'].update_db_subnet(
+            context, subnet['id'], subnet, old_pools)
 
         mocks['driver'].get_subnet_request_factory.assert_called_once_with()
         subnet_factory.get_request.assert_called_once_with(context,
@@ -724,13 +762,16 @@ class TestDbBasePluginIpam(test_db_base.NeutronDbPluginV2TestCase):
 
     @mock.patch('neutron.ipam.driver.Pool')
     def test_update_db_subnet_unchanged_pools(self, pool_mock):
-        old_pools = [netaddr.IPRange('192.1.1.2', '192.1.1.254')]
+        old_pools = [{'start': '192.1.1.2', 'end': '192.1.1.254'}]
+        context = self.admin_context
         subnet = {'id': uuidutils.generate_uuid(),
-                  'network_id': uuidutils.generate_uuid(),
-                  'cidr': '192.1.1.0/24',
+                  'ip_version': constants.IP_VERSION_4,
+                  'cidr': netaddr.IPNetwork('192.1.1.0/24'),
                   'ipv6_address_mode': None,
                   'ipv6_ra_mode': None}
         subnet_with_pools = subnet.copy()
+        subnet_obj = obj_subnet.Subnet(context, **subnet_with_pools)
+        subnet_obj.create()
         subnet_with_pools['allocation_pools'] = old_pools
         # if subnet has no allocation pools set, then old pools has to
         # be added to subnet dict passed to request factory
@@ -739,17 +780,25 @@ class TestDbBasePluginIpam(test_db_base.NeutronDbPluginV2TestCase):
 
     @mock.patch('neutron.ipam.driver.Pool')
     def test_update_db_subnet_new_pools(self, pool_mock):
-        old_pools = [netaddr.IPRange('192.1.1.2', '192.1.1.254')]
+        old_pools = [{'start': '192.1.1.2', 'end': '192.1.1.254'}]
+        context = self.admin_context
         subnet = {'id': uuidutils.generate_uuid(),
-                  'network_id': uuidutils.generate_uuid(),
-                  'cidr': '192.1.1.0/24',
-                  'allocation_pools': [
-                      netaddr.IPRange('192.1.1.10', '192.1.1.254')],
+                  'ip_version': constants.IP_VERSION_4,
+                  'cidr': netaddr.IPNetwork('192.1.1.0/24'),
                   'ipv6_address_mode': None,
                   'ipv6_ra_mode': None}
         # make a copy of subnet for validation, since update_subnet changes
         # incoming subnet dict
         expected_subnet = subnet.copy()
+        subnet_obj = obj_subnet.Subnet(context, **subnet)
+        subnet_obj.create()
+        subnet['allocation_pools'] = [
+            netaddr.IPRange('192.1.1.10', '192.1.1.254')]
+        expected_subnet = subnet.copy()
+        obj_subnet.IPAllocationPool(context,
+                                    subnet_id=subnet['id'],
+                                    start='192.1.1.10',
+                                    end='192.1.1.254').create()
         # validate that subnet passed to request factory is the same as
         # incoming one, i.e. new pools in it are not overwritten by old pools
         self._test_update_db_subnet(pool_mock, subnet, expected_subnet,
@@ -762,21 +811,23 @@ class TestDbBasePluginIpam(test_db_base.NeutronDbPluginV2TestCase):
         mocks['ipam'] = ipam_pluggable_backend.IpamPluggableBackend()
 
         new_port = {'fixed_ips': [{'ip_address': '192.168.1.20',
-                                   'subnet_id': 'some-id'},
+                                   'subnet_id': uuidutils.generate_uuid()},
                                   {'ip_address': '192.168.1.50',
-                                   'subnet_id': 'some-id'}]}
-        db_port = models_v2.Port(id='id', network_id='some-net-id')
+                                   'subnet_id': uuidutils.generate_uuid()}]}
+        db_port = port_obj.Port(context,
+                                id=uuidutils.generate_uuid(),
+                                network_id=uuidutils.generate_uuid())
         old_port = {'fixed_ips': [{'ip_address': '192.168.1.10',
-                                   'subnet_id': 'some-id'},
+                                   'subnet_id': uuidutils.generate_uuid()},
                                   {'ip_address': '192.168.1.50',
-                                   'subnet_id': 'some-id'}]}
+                                   'subnet_id': uuidutils.generate_uuid()}]}
         changes = mocks['ipam'].Changes(
             add=[{'ip_address': '192.168.1.20',
-                  'subnet_id': 'some-id'}],
+                  'subnet_id': uuidutils.generate_uuid()}],
             original=[{'ip_address': '192.168.1.50',
-                       'subnet_id': 'some-id'}],
+                       'subnet_id': uuidutils.generate_uuid()}],
             remove=[{'ip_address': '192.168.1.10',
-                     'subnet_id': 'some-id'}])
+                     'subnet_id': uuidutils.generate_uuid()}])
         mocks['ipam']._delete_ip_allocation = mock.Mock()
         mocks['ipam']._make_port_dict = mock.Mock(return_value=old_port)
         mocks['ipam']._update_ips_for_port = mock.Mock(return_value=changes)
@@ -786,18 +837,19 @@ class TestDbBasePluginIpam(test_db_base.NeutronDbPluginV2TestCase):
         mocks['ipam']._ipam_deallocate_ips = mock.Mock(side_effect=ValueError)
         mocks['ipam']._ipam_allocate_ips = mock.Mock(side_effect=ValueError)
 
-        # Validate original exception (DBDeadlock) is not overriden by
+        # Validate original exception (DBDeadlock) is not overridden by
         # exception raised on rollback (ValueError)
-        self.assertRaises(db_exc.DBDeadlock,
-                          mocks['ipam'].update_port_with_ips,
-                          context,
-                          None,
-                          db_port,
-                          new_port,
-                          mock.Mock())
-        mocks['ipam']._ipam_deallocate_ips.assert_called_once_with(
-            context, mocks['driver'], db_port,
-            changes.add, revert_on_fail=False)
+        with mock.patch.object(port_obj.IPAllocation, 'create'):
+            self.assertRaises(db_exc.DBDeadlock,
+                              mocks['ipam'].update_port_with_ips,
+                              context,
+                              None,
+                              db_port,
+                              new_port,
+                              mock.Mock())
+            mocks['ipam']._ipam_deallocate_ips.assert_called_once_with(
+                context, mocks['driver'], db_port,
+                changes.add, revert_on_fail=False)
         mocks['ipam']._ipam_allocate_ips.assert_called_once_with(
             context, mocks['driver'], db_port,
             changes.remove, revert_on_fail=False)
@@ -860,15 +912,16 @@ class TestRollback(test_db_base.NeutronDbPluginV2TestCase):
 
         Backend = ipam_pluggable_backend.IpamPluggableBackend
         with mock.patch.object(Backend, '_store_ip_allocation', wraps=store),\
-            mock.patch.object(Backend, '_safe_rollback', wraps=rollback),\
-            mock.patch.object(Backend, '_allocate_ips_for_port', wraps=alloc):
-                # Create port with two addresses. The wrapper lets one succeed
-                # then simulates race for the second to trigger IPAM rollback.
-                response = self._create_port(
-                    self.fmt,
-                    net_id=net['network']['id'],
-                    fixed_ips=[{'subnet_id': subnet1['subnet']['id']},
-                               {'subnet_id': subnet2['subnet']['id']}])
+                mock.patch.object(Backend, '_safe_rollback', wraps=rollback),\
+                mock.patch.object(Backend, '_allocate_ips_for_port',
+                                  wraps=alloc):
+            # Create port with two addresses. The wrapper lets one succeed
+            # then simulates race for the second to trigger IPAM rollback.
+            response = self._create_port(
+                self.fmt,
+                net_id=net['network']['id'],
+                fixed_ips=[{'subnet_id': subnet1['subnet']['id']},
+                           {'subnet_id': subnet2['subnet']['id']}])
 
         # When all goes well, retry kicks in and the operation is successful.
         self.assertEqual(webob.exc.HTTPCreated.code, response.status_int)
