@@ -14,29 +14,32 @@
 # limitations under the License.
 
 import contextlib
+import copy
 
 import mock
 from neutron_lib.api import validators
 from neutron_lib import constants as const
+from neutron_lib import context
+from neutron_lib.db import api as db_api
+from neutron_lib.db import constants as db_const
+from neutron_lib import exceptions
+from neutron_lib.plugins import directory
 from oslo_config import cfg
 import oslo_db.exception as exc
-import six
 import testtools
 import webob.exc
 
-from neutron.api.v2 import attributes as attr
-from neutron.common import exceptions as n_exc
-from neutron import context
 from neutron.db import db_base_plugin_v2
 from neutron.db import securitygroups_db
 from neutron.extensions import securitygroup as ext_sg
 from neutron.extensions import standardattrdescription
-from neutron import manager
 from neutron.tests import base
 from neutron.tests.unit.db import test_db_base_plugin_v2
 
 DB_PLUGIN_KLASS = ('neutron.tests.unit.extensions.test_securitygroup.'
                    'SecurityGroupTestPlugin')
+LONG_NAME_OK = 'x' * (db_const.NAME_FIELD_SIZE)
+LONG_NAME_NG = 'x' * (db_const.NAME_FIELD_SIZE + 1)
 
 
 class SecurityGroupTestExtensionManager(object):
@@ -52,12 +55,11 @@ class SecurityGroupTestExtensionManager(object):
                 ext_sg.RESOURCE_ATTRIBUTE_MAP[ext_sg.SECURITYGROUPRULES])
             sg_rule_attr_desc = ext_res[ext_sg.SECURITYGROUPRULES]
             existing_sg_rule_attr_map.update(sg_rule_attr_desc)
-        # Add the resources to the global attribute map
-        # This is done here as the setup process won't
-        # initialize the main API router which extends
-        # the global attribute map
-        attr.RESOURCE_ATTRIBUTE_MAP.update(
-            ext_sg.RESOURCE_ATTRIBUTE_MAP)
+        if ext_sg.SECURITYGROUPS in ext_res:
+            existing_sg_attr_map = (
+                ext_sg.RESOURCE_ATTRIBUTE_MAP[ext_sg.SECURITYGROUPS])
+            sg_attr_desc = ext_res[ext_sg.SECURITYGROUPS]
+            existing_sg_attr_map.update(sg_attr_desc)
         return ext_sg.Securitygroup.get_resources()
 
     def get_actions(self):
@@ -69,14 +71,16 @@ class SecurityGroupTestExtensionManager(object):
 
 class SecurityGroupsTestCase(test_db_base_plugin_v2.NeutronDbPluginV2TestCase):
 
-    def _create_security_group(self, fmt, name, description, **kwargs):
-
+    def _build_security_group(self, name, description, **kwargs):
         data = {
             'security_group': {
                 'name': name,
-                'tenant_id': kwargs.get('tenant_id',
-                                        test_db_base_plugin_v2.TEST_TENANT_ID),
+                'tenant_id': kwargs.get(
+                    'tenant_id', test_db_base_plugin_v2.TEST_TENANT_ID),
                 'description': description}}
+        return data
+
+    def _create_security_group_response(self, fmt, data, **kwargs):
         security_group_req = self.new_create_request('security-groups', data,
                                                      fmt)
         if (kwargs.get('set_context') and 'tenant_id' in kwargs):
@@ -85,6 +89,10 @@ class SecurityGroupsTestCase(test_db_base_plugin_v2.NeutronDbPluginV2TestCase):
                 context.Context('', kwargs['tenant_id']))
         return security_group_req.get_response(self.ext_api)
 
+    def _create_security_group(self, fmt, name, description, **kwargs):
+        data = self._build_security_group(name, description, **kwargs)
+        return self._create_security_group_response(fmt, data, **kwargs)
+
     def _build_security_group_rule(
             self, security_group_id, direction, proto,
             port_range_min=None, port_range_max=None,
@@ -120,6 +128,10 @@ class SecurityGroupsTestCase(test_db_base_plugin_v2.NeutronDbPluginV2TestCase):
             # create a specific auth context for this request
             security_group_rule_req.environ['neutron.context'] = (
                 context.Context('', kwargs['tenant_id']))
+        elif kwargs.get('admin_context'):
+            security_group_rule_req.environ['neutron.context'] = (
+                context.Context(user_id='admin', tenant_id='admin-tenant',
+                is_admin=True))
         return security_group_rule_req.get_response(self.ext_api)
 
     def _make_security_group(self, fmt, name, description, **kwargs):
@@ -169,15 +181,15 @@ class SecurityGroupsTestCase(test_db_base_plugin_v2.NeutronDbPluginV2TestCase):
 
         for r in res['security_group_rules']:
             if (r['direction'] == 'egress' and not r['port_range_max'] and
-                    not r['port_range_min'] and not r['protocol']
-                    and not r['remote_ip_prefix']):
+                    not r['port_range_min'] and not r['protocol'] and
+                    not r['remote_ip_prefix']):
                 self._delete('security-group-rules', r['id'])
 
     def _assert_sg_rule_has_kvs(self, security_group_rule, expected_kvs):
         """Asserts that the sg rule has expected key/value pairs passed
            in as expected_kvs dictionary
         """
-        for k, v in six.iteritems(expected_kvs):
+        for k, v in expected_kvs.items():
             self.assertEqual(security_group_rule[k], v)
 
 
@@ -197,8 +209,7 @@ class SecurityGroupTestPlugin(db_base_plugin_v2.NeutronDbPluginV2,
         default_sg = self._ensure_default_security_group(context, tenant_id)
         if not validators.is_attr_set(port['port'].get(ext_sg.SECURITYGROUPS)):
             port['port'][ext_sg.SECURITYGROUPS] = [default_sg]
-        session = context.session
-        with session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             sgids = self._get_security_groups_on_port(context, port)
             port = super(SecurityGroupTestPlugin, self).create_port(context,
                                                                     port)
@@ -207,8 +218,7 @@ class SecurityGroupTestPlugin(db_base_plugin_v2.NeutronDbPluginV2,
         return port
 
     def update_port(self, context, id, port):
-        session = context.session
-        with session.begin(subtransactions=True):
+        with db_api.CONTEXT_WRITER.using(context):
             if ext_sg.SECURITYGROUPS in port['port']:
                 port['port'][ext_sg.SECURITYGROUPS] = (
                     self._get_security_groups_on_port(context, port))
@@ -240,11 +250,16 @@ class SecurityGroupTestPlugin(db_base_plugin_v2.NeutronDbPluginV2,
 
 class SecurityGroupDBTestCase(SecurityGroupsTestCase):
     def setUp(self, plugin=None, ext_mgr=None):
+        self._backup = copy.deepcopy(ext_sg.RESOURCE_ATTRIBUTE_MAP)
+        self.addCleanup(self._restore)
         plugin = plugin or DB_PLUGIN_KLASS
         ext_mgr = ext_mgr or SecurityGroupTestExtensionManager()
         super(SecurityGroupDBTestCase,
               self).setUp(plugin=plugin, ext_mgr=ext_mgr)
 
+    def _restore(self):
+        ext_sg.RESOURCE_ATTRIBUTE_MAP = self._backup
+
 
 class TestSecurityGroups(SecurityGroupDBTestCase):
     def test_create_security_group(self):
@@ -284,6 +299,16 @@ class TestSecurityGroups(SecurityGroupDBTestCase):
                     'port_range_min': None}
         self._assert_sg_rule_has_kvs(v6_rule, expected)
 
+    def test_create_security_group_bulk(self):
+        rule1 = self._build_security_group("sg_1", "sec_grp_1")
+        rule2 = self._build_security_group("sg_2", "sec_grp_2")
+        rules = {'security_groups': [rule1['security_group'],
+                                     rule2['security_group']]}
+        res = self._create_security_group_response(self.fmt, rules)
+        ret = self.deserialize(self.fmt, res)
+        self.assertEqual(webob.exc.HTTPCreated.code, res.status_int)
+        self.assertEqual(2, len(ret['security_groups']))
+
     def test_skip_duplicate_default_sg_error(self):
         num_called = [0]
         original_func = self.plugin.create_security_group
@@ -291,7 +316,7 @@ class TestSecurityGroups(SecurityGroupDBTestCase):
         def side_effect(context, security_group, default_sg):
             # can't always raise, or create_security_group will hang
             self.assertTrue(default_sg)
-            self.assertTrue(num_called[0] < 2)
+            self.assertLess(num_called[0], 2)
             num_called[0] += 1
             ret = original_func(context, security_group, default_sg)
             if num_called[0] == 1:
@@ -362,6 +387,39 @@ class TestSecurityGroups(SecurityGroupDBTestCase):
             self.assertEqual(data['security_group']['description'],
                              res['security_group']['description'])
 
+    def test_update_security_group_with_max_name_length(self):
+        with self.security_group() as sg:
+            data = {'security_group': {'name': LONG_NAME_OK,
+                                       'description': 'new_desc'}}
+            req = self.new_update_request('security-groups',
+                                          data,
+                                          sg['security_group']['id'])
+            res = self.deserialize(self.fmt, req.get_response(self.ext_api))
+            self.assertEqual(data['security_group']['name'],
+                             res['security_group']['name'])
+            self.assertEqual(data['security_group']['description'],
+                             res['security_group']['description'])
+
+    def test_update_security_group_with_too_long_name(self):
+        with self.security_group() as sg:
+            data = {'security_group': {'name': LONG_NAME_NG,
+                                       'description': 'new_desc'}}
+            req = self.new_update_request('security-groups',
+                                          data,
+                                          sg['security_group']['id'])
+            res = req.get_response(self.ext_api)
+            self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)
+
+    def test_update_security_group_with_boolean_type_name(self):
+        with self.security_group() as sg:
+            data = {'security_group': {'name': True,
+                                       'description': 'new_desc'}}
+            req = self.new_update_request('security-groups',
+                                          data,
+                                          sg['security_group']['id'])
+            res = req.get_response(self.ext_api)
+            self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)
+
     def test_check_default_security_group_description(self):
         with self.network():
             res = self.new_list_request('security-groups')
@@ -389,6 +447,24 @@ class TestSecurityGroups(SecurityGroupDBTestCase):
         self.deserialize(self.fmt, res)
         self.assertEqual(webob.exc.HTTPConflict.code, res.status_int)
 
+    def test_create_security_group_with_max_name_length(self):
+        description = 'my webservers'
+        res = self._create_security_group(self.fmt, LONG_NAME_OK, description)
+        self.deserialize(self.fmt, res)
+        self.assertEqual(webob.exc.HTTPCreated.code, res.status_int)
+
+    def test_create_security_group_with_too_long_name(self):
+        description = 'my webservers'
+        res = self._create_security_group(self.fmt, LONG_NAME_NG, description)
+        self.deserialize(self.fmt, res)
+        self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)
+
+    def test_create_security_group_with_boolean_type_name(self):
+        description = 'my webservers'
+        res = self._create_security_group(self.fmt, True, description)
+        self.deserialize(self.fmt, res)
+        self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)
+
     def test_list_security_groups(self):
         with self.security_group(name='sg1', description='sg') as v1,\
                 self.security_group(name='sg2', description='sg') as v2,\
@@ -470,7 +546,7 @@ class TestSecurityGroups(SecurityGroupDBTestCase):
         test_addr = {'192.168.1.1/24': 'IPv6',
                      '2001:db8:1234::/48': 'IPv4',
                      '192.168.2.1/24': 'BadEthertype'}
-        for remote_ip_prefix, ethertype in six.iteritems(test_addr):
+        for remote_ip_prefix, ethertype in test_addr.items():
             with self.security_group(name, description) as sg:
                 sg_id = sg['security_group']['id']
                 rule = self._build_security_group_rule(
@@ -532,6 +608,46 @@ class TestSecurityGroups(SecurityGroupDBTestCase):
             self.deserialize(self.fmt, res)
             self.assertEqual(webob.exc.HTTPCreated.code, res.status_int)
 
+    def test_create_security_group_rule_protocol_as_number_with_port_bad(self):
+        # When specifying ports, neither can be None
+        name = 'webservers'
+        description = 'my webservers'
+        with self.security_group(name, description) as sg:
+            security_group_id = sg['security_group']['id']
+            protocol = 6
+            rule = self._build_security_group_rule(
+                security_group_id, 'ingress', protocol, '70', None)
+            res = self._create_security_group_rule(self.fmt, rule)
+            self.deserialize(self.fmt, res)
+            self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)
+
+    def test_create_security_group_rule_protocol_as_number_range(self):
+        # This is a SG rule with a port range, but treated as a single
+        # port since min/max are the same.
+        name = 'webservers'
+        description = 'my webservers'
+        with self.security_group(name, description) as sg:
+            security_group_id = sg['security_group']['id']
+            protocol = 6
+            rule = self._build_security_group_rule(
+                security_group_id, 'ingress', protocol, '70', '70')
+            res = self._create_security_group_rule(self.fmt, rule)
+            self.deserialize(self.fmt, res)
+            self.assertEqual(webob.exc.HTTPCreated.code, res.status_int)
+
+    def test_create_security_group_rule_protocol_as_number_port_bad(self):
+        # Only certain protocols support a SG rule with a port
+        name = 'webservers'
+        description = 'my webservers'
+        with self.security_group(name, description) as sg:
+            security_group_id = sg['security_group']['id']
+            protocol = 111
+            rule = self._build_security_group_rule(
+                security_group_id, 'ingress', protocol, '70', '70')
+            res = self._create_security_group_rule(self.fmt, rule)
+            self.deserialize(self.fmt, res)
+            self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)
+
     def test_create_security_group_rule_case_insensitive(self):
         name = 'webservers'
         description = 'my webservers'
@@ -588,8 +704,89 @@ class TestSecurityGroups(SecurityGroupDBTestCase):
                 for k, v, in keys:
                     self.assertEqual(sg_rule[0][k], v)
 
+    def test_get_security_group_empty_rules(self):
+        name = 'webservers'
+        description = 'my webservers'
+        with self.security_group(name, description) as sg:
+            remote_group_id = sg['security_group']['id']
+
+            self._delete_default_security_group_egress_rules(
+                remote_group_id)
+
+            res = self.new_show_request('security-groups', remote_group_id)
+            group = self.deserialize(
+                self.fmt, res.get_response(self.ext_api))
+
+            sg_rule = group['security_group']['security_group_rules']
+            self.assertEqual(group['security_group']['id'],
+                             remote_group_id)
+            self.assertEqual(0, len(sg_rule))
+
+    def test_get_security_group_empty_rules_id_only(self):
+        name = 'webservers'
+        description = 'my webservers'
+        with self.security_group(name, description) as sg:
+            remote_group_id = sg['security_group']['id']
+
+            self._delete_default_security_group_egress_rules(
+                remote_group_id)
+
+            res = self.new_show_request('security-groups', remote_group_id,
+                                        fields=['id'])
+            group = self.deserialize(
+                self.fmt, res.get_response(self.ext_api))
+
+            secgroup = group['security_group']
+            self.assertFalse('security_group_rules' in secgroup)
+            self.assertEqual(group['security_group']['id'],
+                             remote_group_id)
+
+    # This test case checks that admins from a different tenant can add rules
+    # as themselves. This is an odd behavior, with some weird GET semantics,
+    # but this test is checking that we don't break that old behavior, at least
+    # until we make a conscious choice to do so.
+    def test_create_security_group_rules_admin_tenant(self):
+        name = 'webservers'
+        description = 'my webservers'
+        with self.security_group(name, description) as sg:
+            # Add a couple normal rules
+            rule = self._build_security_group_rule(
+                sg['security_group']['id'], "ingress", const.PROTO_NAME_TCP,
+                port_range_min=22, port_range_max=22,
+                remote_ip_prefix="10.0.0.0/24",
+                ethertype=const.IPv4)
+            self._make_security_group_rule(self.fmt, rule)
+
+            rule = self._build_security_group_rule(
+                sg['security_group']['id'], "ingress", const.PROTO_NAME_TCP,
+                port_range_min=22, port_range_max=22,
+                remote_ip_prefix="10.0.1.0/24",
+                ethertype=const.IPv4)
+            self._make_security_group_rule(self.fmt, rule)
+
+            # Let's add a rule as admin, with a different tenant_id. The
+            # results of this call are arguably a bug, but it is past behavior.
+            rule = self._build_security_group_rule(
+                sg['security_group']['id'], "ingress", const.PROTO_NAME_TCP,
+                port_range_min=22, port_range_max=22,
+                remote_ip_prefix="10.0.2.0/24",
+                ethertype=const.IPv4,
+                tenant_id='admin-tenant')
+            self._make_security_group_rule(self.fmt, rule, admin_context=True)
+
+            # Now, let's make sure all the rules are there, with their odd
+            # tenant_id behavior.
+            res = self.new_list_request('security-groups')
+            sgs = self.deserialize(self.fmt, res.get_response(self.ext_api))
+            for sg in sgs['security_groups']:
+                if sg['name'] == "webservers":
+                    rules = sg['security_group_rules']
+                    self.assertEqual(len(rules), 5)
+                    self.assertNotEqual(rules[3]['tenant_id'], 'admin-tenant')
+                    self.assertEqual(rules[4]['tenant_id'], 'admin-tenant')
+
     def test_get_security_group_on_port_from_wrong_tenant(self):
-        plugin = manager.NeutronManager.get_plugin()
+        plugin = directory.get_plugin()
         if not hasattr(plugin, '_get_security_groups_on_port'):
             self.skipTest("plugin doesn't use the mixin with this method")
         neutron_context = context.get_admin_context()
@@ -1364,6 +1561,25 @@ class TestSecurityGroups(SecurityGroupDBTestCase):
                                      res['port'].get(ext_sg.SECURITYGROUPS))
                     self._delete('ports', port['port']['id'])
 
+    def test_update_port_with_invalid_type_in_security_groups_param(self):
+        with self.network() as n:
+            with self.subnet(n):
+                with self.security_group() as sg:
+                    res = self._create_port(self.fmt, n['network']['id'],
+                                            security_groups=(
+                                                [sg['security_group']['id']]))
+                    port = self.deserialize(self.fmt, res)
+
+                    data = {'port': {'fixed_ips': port['port']['fixed_ips'],
+                                     'name': port['port']['name'],
+                                     'security_groups': True}}
+
+                    req = self.new_update_request('ports', data,
+                                                  port['port']['id'])
+                    res = req.get_response(self.api)
+                    self.assertEqual(webob.exc.HTTPBadRequest.code,
+                                     res.status_int)
+
     def test_create_port_with_bad_security_group(self):
         with self.network() as n:
             with self.subnet(n):
@@ -1373,6 +1589,14 @@ class TestSecurityGroups(SecurityGroupDBTestCase):
                 self.deserialize(self.fmt, res)
                 self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)
 
+    def test_create_port_with_invalid_type_in_security_groups_param(self):
+        with self.network() as n:
+            with self.subnet(n):
+                res = self._create_port(self.fmt, n['network']['id'],
+                                        security_groups=True)
+
+                self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)
+
     def test_create_delete_security_group_port_in_use(self):
         with self.network() as n:
             with self.subnet(n):
@@ -1413,7 +1637,7 @@ class TestSecurityGroups(SecurityGroupDBTestCase):
     def test_create_security_group_rule_bulk_emulated(self):
         real_has_attr = hasattr
 
-        #ensures the API choose the emulation code path
+        # ensures the API choose the emulation code path
         def fakehasattr(item, attr):
             if attr.endswith('__native_bulk_support'):
                 return False
@@ -1484,7 +1708,7 @@ class TestSecurityGroups(SecurityGroupDBTestCase):
     def test_create_security_group_rule_duplicate_rule_in_post_emulated(self):
         real_has_attr = hasattr
 
-        #ensures the API choose the emulation code path
+        # ensures the API choose the emulation code path
         def fakehasattr(item, attr):
             if attr.endswith('__native_bulk_support'):
                 return False
@@ -1521,7 +1745,7 @@ class TestSecurityGroups(SecurityGroupDBTestCase):
     def test_create_security_group_rule_duplicate_rule_db_emulated(self):
         real_has_attr = hasattr
 
-        #ensures the API choose the emulation code path
+        # ensures the API choose the emulation code path
         def fakehasattr(item, attr):
             if attr.endswith('__native_bulk_support'):
                 return False
@@ -1674,13 +1898,13 @@ class TestConvertIPPrefixToCIDR(base.BaseTestCase):
 
     def test_convert_bad_ip_prefix_to_cidr(self):
         for val in ['bad_ip', 256, "2001:db8:a::123/129"]:
-            self.assertRaises(n_exc.InvalidCIDR,
+            self.assertRaises(exceptions.InvalidCIDR,
                               ext_sg.convert_ip_prefix_to_cidr, val)
         self.assertIsNone(ext_sg.convert_ip_prefix_to_cidr(None))
 
     def test_convert_ip_prefix_no_netmask_to_cidr(self):
         addr = {'10.1.2.3': '32', 'fe80::2677:3ff:fe7d:4c': '128'}
-        for k, v in six.iteritems(addr):
+        for k, v in addr.items():
             self.assertEqual(ext_sg.convert_ip_prefix_to_cidr(k),
                              '%s/%s' % (k, v))
 
